# Generated by generateOTTagsModule.py
# Scraped from:
# https://docs.microsoft.com/en-us/typography/opentype/spec/featurelist
# https://docs.microsoft.com/en-us/typography/opentype/spec/scripttags
# https://docs.microsoft.com/en-us/typography/opentype/spec/languagetags


__all__ = ['features', 'scripts', 'languages']


features = {
    # tag, friendly name, documentation URL
    'aalt': ('Access All Alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#aalt'),
    'abvf': ('Above-base Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#abvf'),
    'abvm': ('Above-base Mark Positioning', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#abvm'),
    'abvs': ('Above-base Substitutions', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#abvs'),
    'afrc': ('Alternative Fractions', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#afrc'),
    'akhn': ('Akhands', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#akhn'),
    'blwf': ('Below-base Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#blwf'),
    'blwm': ('Below-base Mark Positioning', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#blwm'),
    'blws': ('Below-base Substitutions', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#blws'),
    'calt': ('Contextual Alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#calt'),
    'case': ('Case-Sensitive Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#case'),
    'ccmp': ('Glyph Composition / Decomposition', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#ccmp'),
    'cfar': ('Conjunct Form After Ro', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cfar'),
    'cjct': ('Conjunct Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cjct'),
    'clig': ('Contextual Ligatures', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#clig'),
    'cpct': ('Centered CJK Punctuation', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cpct'),
    'cpsp': ('Capital Spacing', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cpsp'),
    'cswh': ('Contextual Swash', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cswh'),
    'curs': ('Cursive Positioning', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#curs'),
    'cv01': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv02': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv03': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv04': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv05': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv06': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv07': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv08': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv09': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv10': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv11': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv12': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv13': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv14': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv15': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv16': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv17': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv18': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv19': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv20': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv21': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv22': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv23': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv24': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv25': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv26': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv27': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv28': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv29': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv30': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv31': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv32': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv33': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv34': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv35': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv36': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv37': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv38': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv39': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv40': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv41': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv42': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv43': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv44': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv45': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv46': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv47': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv48': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv49': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv50': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv51': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv52': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv53': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv54': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv55': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv56': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv57': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv58': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv59': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv60': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv61': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv62': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv63': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv64': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv65': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv66': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv67': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv68': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv69': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv70': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv71': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv72': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv73': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv74': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv75': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv76': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv77': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv78': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv79': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv80': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv81': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv82': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv83': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv84': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv85': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv86': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv87': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv88': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv89': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv90': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv91': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv92': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv93': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv94': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv95': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv96': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv97': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv98': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'cv99': ('Character Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#cv01-cv99'),
    'c2pc': ('Petite Capitals From Capitals', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#c2pc'),
    'c2sc': ('Small Capitals From Capitals', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#c2sc'),
    'dist': ('Distances', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#dist'),
    'dlig': ('Discretionary Ligatures', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#dlig'),
    'dnom': ('Denominators', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#dnom'),
    'dtls': ('Dotless Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#dtls'),
    'expt': ('Expert Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ae#expt'),
    'falt': ('Final Glyph on Line Alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#falt'),
    'fin2': ('Terminal Forms #2', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#fin2'),
    'fin3': ('Terminal Forms #3', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#fin3'),
    'fina': ('Terminal Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#fina'),
    'flac': ('Flattened accent forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#flac'),
    'frac': ('Fractions', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#frac'),
    'fwid': ('Full Widths', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#fwid'),
    'half': ('Half Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#half'),
    'haln': ('Halant Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#haln'),
    'halt': ('Alternate Half Widths', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#halt'),
    'hist': ('Historical Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#hist'),
    'hkna': ('Horizontal Kana Alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#hkna'),
    'hlig': ('Historical Ligatures', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#hlig'),
    'hngl': ('Hangul', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#hngl'),
    'hojo': ('Hojo Kanji Forms (JIS X 0212-1990 Kanji Forms)', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#hojo'),
    'hwid': ('Half Widths', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#hwid'),
    'init': ('Initial Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#init'),
    'isol': ('Isolated Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#isol'),
    'ital': ('Italics', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#ital'),
    'jalt': ('Justification Alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#jalt'),
    'jp78': ('JIS78 Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#jp78'),
    'jp83': ('JIS83 Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#jp83'),
    'jp90': ('JIS90 Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#jp90'),
    'jp04': ('JIS2004 Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_fj#jp04'),
    'kern': ('Kerning', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#kern'),
    'lfbd': ('Left Bounds', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#lfbd'),
    'liga': ('Standard Ligatures', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#liga'),
    'ljmo': ('Leading Jamo Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#ljmo'),
    'lnum': ('Lining Figures', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#lnum'),
    'locl': ('Localized Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#locl'),
    'ltra': ('Left-to-right alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#ltra'),
    'ltrm': ('Left-to-right mirrored forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#ltrm'),
    'mark': ('Mark Positioning', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#mark'),
    'med2': ('Medial Forms #2', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#med2'),
    'medi': ('Medial Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#medi'),
    'mgrk': ('Mathematical Greek', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#mgrk'),
    'mkmk': ('Mark to Mark Positioning', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#mkmk'),
    'mset': ('Mark Positioning via Substitution', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#mset'),
    'nalt': ('Alternate Annotation Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#nalt'),
    'nlck': ('NLC Kanji Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#nlck'),
    'nukt': ('Nukta Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#nukt'),
    'numr': ('Numerators', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#numr'),
    'onum': ('Oldstyle Figures', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#onum'),
    'opbd': ('Optical Bounds', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#opbd'),
    'ordn': ('Ordinals', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#ordn'),
    'ornm': ('Ornaments', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_ko#ornm'),
    'palt': ('Proportional Alternate Widths', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#palt'),
    'pcap': ('Petite Capitals', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#pcap'),
    'pkna': ('Proportional Kana', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#pkna'),
    'pnum': ('Proportional Figures', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#pnum'),
    'pref': ('Pre-Base Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#pref'),
    'pres': ('Pre-base Substitutions', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#pres'),
    'pstf': ('Post-base Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#pstf'),
    'psts': ('Post-base Substitutions', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#psts'),
    'pwid': ('Proportional Widths', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#pwid'),
    'qwid': ('Quarter Widths', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#qwid'),
    'rand': ('Randomize', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#rand'),
    'rclt': ('Required Contextual Alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#rclt'),
    'rkrf': ('Rakar Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#rkrf'),
    'rlig': ('Required Ligatures', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#rlig'),
    'rphf': ('Reph Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#rphf'),
    'rtbd': ('Right Bounds', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#rtbd'),
    'rtla': ('Right-to-left alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#rtla'),
    'rtlm': ('Right-to-left mirrored forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#rtlm'),
    'ruby': ('Ruby Notation Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ruby'),
    'rvrn': ('Required Variation Alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#rvrn'),
    'salt': ('Stylistic Alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#salt'),
    'sinf': ('Scientific Inferiors', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#sinf'),
    'size': ('Optical size', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#size'),
    'smcp': ('Small Capitals', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#smcp'),
    'smpl': ('Simplified Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#smpl'),
    'ss01': ('Stylistic Set 1', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss02': ('Stylistic Set 2', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss03': ('Stylistic Set 3', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss04': ('Stylistic Set 4', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss05': ('Stylistic Set 5', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss06': ('Stylistic Set 6', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss07': ('Stylistic Set 7', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss08': ('Stylistic Set 8', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss09': ('Stylistic Set 9', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss10': ('Stylistic Set 10', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss11': ('Stylistic Set 11', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss12': ('Stylistic Set 12', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss13': ('Stylistic Set 13', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss14': ('Stylistic Set 14', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss15': ('Stylistic Set 15', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss16': ('Stylistic Set 16', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss17': ('Stylistic Set 17', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss18': ('Stylistic Set 18', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss19': ('Stylistic Set 19', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ss20': ('Stylistic Set 20', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssxx'),
    'ssty': ('Math script style alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#ssty'),
    'stch': ('Stretching Glyph Decomposition', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#stch'),
    'subs': ('Subscript', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#subs'),
    'sups': ('Superscript', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#sups'),
    'swsh': ('Swash', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#swsh'),
    'titl': ('Titling', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#titl'),
    'tjmo': ('Trailing Jamo Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#tjmo'),
    'tnam': ('Traditional Name Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#tnam'),
    'tnum': ('Tabular Figures', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#tnum'),
    'trad': ('Traditional Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#trad'),
    'twid': ('Third Widths', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_pt#twid'),
    'unic': ('Unicase', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#unic'),
    'valt': ('Alternate Vertical Metrics', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#valt'),
    'vatu': ('Vattu Variants', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#vatu'),
    'vert': ('Vertical Writing', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#vert'),
    'vhal': ('Alternate Vertical Half Metrics', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#vhal'),
    'vjmo': ('Vowel Jamo Forms', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#vjmo'),
    'vkna': ('Vertical Kana Alternates', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#vkna'),
    'vkrn': ('Vertical Kerning', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#vkrn'),
    'vpal': ('Proportional Alternate Vertical Metrics', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#vpal'),
    'vrt2': ('Vertical Alternates and Rotation', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#vrt2'),
    'vrtr': ('Vertical Alternates for Rotation', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#vrtr'),
    'zero': ('Slashed Zero', 'https://docs.microsoft.com/en-us/typography/opentype/spec/features_uz#zero'),
}

scripts = {
    # tag, friendly name
    'adlm': 'Adlam',
    'ahom': 'Ahom',
    'hluw': 'Anatolian Hieroglyphs',
    'arab': 'Arabic',
    'armn': 'Armenian',
    'avst': 'Avestan',
    'bali': 'Balinese',
    'bamu': 'Bamum',
    'bass': 'Bassa Vah',
    'batk': 'Batak',
    'beng': 'Bengali',
    'bng2': 'Bengali v.2',
    'bhks': 'Bhaiksuki',
    'bopo': 'Bopomofo',
    'brah': 'Brahmi',
    'brai': 'Braille',
    'bugi': 'Buginese',
    'buhd': 'Buhid',
    'byzm': 'Byzantine Music',
    'cans': 'Canadian Syllabics',
    'cari': 'Carian',
    'aghb': 'Caucasian Albanian',
    'cakm': 'Chakma',
    'cham': 'Cham',
    'cher': 'Cherokee',
    'hani': 'CJK Ideographic',
    'copt': 'Coptic',
    'cprt': 'Cypriot Syllabary',
    'cyrl': 'Cyrillic',
    'DFLT': 'Default',
    'dsrt': 'Deseret',
    'deva': 'Devanagari',
    'dev2': 'Devanagari v.2',
    'dogr': 'Dogra',
    'dupl': 'Duployan',
    'egyp': 'Egyptian Hieroglyphs',
    'elba': 'Elbasan',
    'ethi': 'Ethiopic',
    'geor': 'Georgian',
    'glag': 'Glagolitic',
    'goth': 'Gothic',
    'gran': 'Grantha',
    'grek': 'Greek',
    'gujr': 'Gujarati',
    'gjr2': 'Gujarati v.2',
    'gong': 'Gunjala Gondi',
    'guru': 'Gurmukhi',
    'gur2': 'Gurmukhi v.2',
    'hang': 'Hangul',
    'jamo': 'Hangul Jamo',
    'rohg': 'Hanifi Rohingya',
    'hano': 'Hanunoo',
    'hatr': 'Hatran',
    'hebr': 'Hebrew',
    'kana': 'Hiragana, Katakana',
    'armi': 'Imperial Aramaic',
    'phli': 'Inscriptional Pahlavi',
    'prti': 'Inscriptional Parthian',
    'java': 'Javanese',
    'kthi': 'Kaithi',
    'knda': 'Kannada',
    'knd2': 'Kannada v.2',
    'kali': 'Kayah Li',
    'khar': 'Kharosthi',
    'khmr': 'Khmer',
    'khoj': 'Khojki',
    'sind': 'Khudawadi',
    'lao ': 'Lao',
    'latn': 'Latin',
    'lepc': 'Lepcha',
    'limb': 'Limbu',
    'lina': 'Linear A',
    'linb': 'Linear B',
    'lisu': 'Lisu (Fraser)',
    'lyci': 'Lycian',
    'lydi': 'Lydian',
    'mahj': 'Mahajani',
    'maka': 'Makasar',
    'mlym': 'Malayalam',
    'mlm2': 'Malayalam v.2',
    'mand': 'Mandaic, Mandaean',
    'mani': 'Manichaean',
    'marc': 'Marchen',
    'gonm': 'Masaram Gondi',
    'math': 'Mathematical Alphanumeric Symbols',
    'medf': 'Medefaidrin (Oberi Okaime, Oberi Ɔkaimɛ)',
    'mtei': 'Meitei Mayek (Meithei, Meetei)',
    'mend': 'Mende Kikakui',
    'merc': 'Meroitic Cursive',
    'mero': 'Meroitic Hieroglyphs',
    'plrd': 'Miao',
    'modi': 'Modi',
    'mong': 'Mongolian',
    'mroo': 'Mro',
    'mult': 'Multani',
    'musc': 'Musical Symbols',
    'mymr': 'Myanmar',
    'mym2': 'Myanmar v.2',
    'nbat': 'Nabataean',
    'newa': 'Newa',
    'talu': 'New Tai Lue',
    'nko ': 'Ko',
    'nshu': 'Nüshu',
    'orya': 'Odia (formerly Oriya)',
    'ory2': 'Odia v.2 (formerly Oriya v.2)',
    'ogam': 'Ogham',
    'olck': 'Ol Chiki',
    'ital': 'Old Italic',
    'hung': 'Old Hungarian',
    'narb': 'Old North Arabian',
    'perm': 'Old Permic',
    'xpeo': 'Old Persian Cuneiform',
    'sogo': 'Old Sogdian',
    'sarb': 'Old South Arabian',
    'orkh': 'Old Turkic, Orkhon Runic',
    'osge': 'Osage',
    'osma': 'Osmanya',
    'hmng': 'Pahawh Hmong',
    'palm': 'Palmyrene',
    'pauc': 'Pau Cin Hau',
    'phag': 'Phags-pa',
    'phnx': 'Phoenician',
    'phlp': 'Psalter Pahlavi',
    'rjng': 'Rejang',
    'runr': 'Runic',
    'samr': 'Samaritan',
    'saur': 'Saurashtra',
    'shrd': 'Sharada',
    'shaw': 'Shavian',
    'sidd': 'Siddham',
    'sgnw': 'Sign Writing',
    'sinh': 'Sinhala',
    'sogd': 'Sogdian',
    'sora': 'Sora Sompeng',
    'soyo': 'Soyombo',
    'xsux': 'Sumero-Akkadian Cuneiform',
    'sund': 'Sundanese',
    'sylo': 'Syloti Nagri',
    'syrc': 'Syriac',
    'tglg': 'Tagalog',
    'tagb': 'Tagbanwa',
    'tale': 'Tai Le',
    'lana': 'Tai Tham (Lanna)',
    'tavt': 'Tai Viet',
    'takr': 'Takri',
    'taml': 'Tamil',
    'tml2': 'Tamil v.2',
    'tang': 'Tangut',
    'telu': 'Telugu',
    'tel2': 'Telugu v.2',
    'thaa': 'Thaana',
    'thai': 'Thai',
    'tibt': 'Tibetan',
    'tfng': 'Tifinagh',
    'tirh': 'Tirhuta',
    'ugar': 'Ugaritic Cuneiform',
    'vai ': 'Vai',
    'wara': 'Warang Citi',
    'yi  ': 'Yi',
    'zanb': 'Zanabazar Square (Zanabazarin Dörböljin Useg, Xewtee Dörböljin Bicig, Horizontal Square Script)',
}

languages = {
    # tag, friendly name, ISO 639 IDs (if applicable)
    'ABA ': ('Abaza', 'abq'),
    'ABK ': ('Abkhazian', 'abk'),
    'ACH ': ('Acholi', 'ach'),
    'ACR ': ('Achi', 'acr'),
    'ADY ': ('Adyghe', 'ady'),
    'AFK ': ('Afrikaans', 'afr'),
    'AFR ': ('Afar', 'aar'),
    'AGW ': ('Agaw', 'ahg'),
    'AIO ': ('Aiton', 'aio'),
    'AKA ': ('Akan', 'aka'),
    'ALS ': ('Alsatian', 'gsw'),
    'ALT ': ('Altai', 'atv', 'alt'),
    'AMH ': ('Amharic', 'amh'),
    'ANG ': ('Anglo-Saxon', 'ang'),
    'APPH': ('Phonetic transcription—Americanist conventions',),
    'ARA ': ('Arabic', 'ara'),
    'ARG ': ('Aragonese', 'arg'),
    'ARI ': ('Aari', 'aiw'),
    'ARK ': ('Rakhine', 'mhv', 'rmz', 'rki'),
    'ASM ': ('Assamese', 'asm'),
    'AST ': ('Asturian', 'ast'),
    'ATH ': ('Athapaskan', 'apk', 'apj', 'apl', 'apm', 'apw', 'nav', 'bea', 'sek', 'bcr', 'caf', 'crx', 'clc', 'gwi', 'haa', 'chp', 'dgr', 'scs', 'xsl', 'srs', 'ing', 'hoi', 'koy', 'hup', 'ktw', 'mvb', 'wlk', 'coq', 'ctc', 'gce', 'tol', 'tuu', 'kkz', 'tgx', 'tht', 'aht', 'tfn', 'taa', 'tau', 'tcb', 'kuu', 'tce', 'ttm', 'txc'),
    'AVR ': ('Avar', 'ava'),
    'AWA ': ('Awadhi', 'awa'),
    'AYM ': ('Aymara', 'aym'),
    'AZB ': ('Torki', 'azb'),
    'AZE ': ('Azerbaijani', 'aze'),
    'BAD ': ('Badaga', 'bfq'),
    'BAD0': ('Banda', 'bad'),
    'BAG ': ('Baghelkhandi', 'bfy'),
    'BAL ': ('Balkar', 'krc'),
    'BAN ': ('Balinese', 'ban'),
    'BAR ': ('Bavarian', 'bar'),
    'BAU ': ('Baulé', 'bci'),
    'BBC ': ('Batak Toba', 'bbc'),
    'BBR ': ('Berber',),
    'BCH ': ('Bench', 'bcq'),
    'BCR ': ('Bible Cree',),
    'BDY ': ('Bandjalang', 'bdy'),
    'BEL ': ('Belarussian', 'bel'),
    'BEM ': ('Bemba', 'bem'),
    'BEN ': ('Bengali', 'ben'),
    'BGC ': ('Haryanvi', 'bgc'),
    'BGQ ': ('Bagri', 'bgq'),
    'BGR ': ('Bulgarian', 'bul'),
    'BHI ': ('Bhili', 'bhi', 'bhb'),
    'BHO ': ('Bhojpuri', 'bho'),
    'BIK ': ('Bikol', 'bik', 'bhk', 'bcl', 'bto', 'cts', 'bln', 'fbl', 'lbl', 'rbl', 'ubl'),
    'BIL ': ('Bilen', 'byn'),
    'BIS ': ('Bislama', 'bis'),
    'BJJ ': ('Kanauji', 'bjj'),
    'BKF ': ('Blackfoot', 'bla'),
    'BLI ': ('Baluchi', 'bal'),
    'BLK ': ('Pa’o Karen', 'blk'),
    'BLN ': ('Balante', 'bjt', 'ble'),
    'BLT ': ('Balti', 'bft'),
    'BMB ': ('Bambara (Bamanankan)', 'bam'),
    'BML ': ('Bamileke',),
    'BOS ': ('Bosnian', 'bos'),
    'BPY ': ('Bishnupriya Manipuri', 'bpy'),
    'BRE ': ('Breton', 'bre'),
    'BRH ': ('Brahui', 'brh'),
    'BRI ': ('Braj Bhasha', 'bra'),
    'BRM ': ('Burmese', 'mya'),
    'BRX ': ('Bodo', 'brx'),
    'BSH ': ('Bashkir', 'bak'),
    'BSK ': ('Burushaski', 'bsk'),
    'BTI ': ('Beti', 'btb', 'beb', 'bum', 'bxp', 'eto', 'ewo', 'mct'),
    'BTS ': ('Batak Simalungun', 'bts'),
    'BUG ': ('Bugis', 'bug'),
    'BYV ': ('Medumba', 'byv'),
    'CAK ': ('Kaqchikel', 'cak'),
    'CAT ': ('Catalan', 'cat'),
    'CBK ': ('Zamboanga Chavacano', 'cbk'),
    'CCHN': ('Chinantec', 'cco', 'chj', 'chq', 'chz', 'cle', 'cnl', 'cnt', 'cpa', 'csa', 'cso', 'cte', 'ctl', 'cuc', 'cvn'),
    'CEB ': ('Cebuano', 'ceb'),
    'CHE ': ('Chechen', 'che'),
    'CHG ': ('Chaha Gurage', 'sgw'),
    'CHH ': ('Chattisgarhi', 'hne'),
    'CHI ': ('Chichewa (Chewa, Nyanja)', 'nya'),
    'CHK ': ('Chukchi', 'ckt'),
    'CHK0': ('Chuukese', 'chk'),
    'CHO ': ('Choctaw', 'cho'),
    'CHP ': ('Chipewyan', 'chp'),
    'CHR ': ('Cherokee', 'chr'),
    'CHA ': ('Chamorro', 'cha'),
    'CHU ': ('Chuvash', 'chv'),
    'CHY ': ('Cheyenne', 'chy'),
    'CGG ': ('Chiga', 'cgg'),
    'CJA ': ('Western Cham', 'cja'),
    'CJM ': ('Eastern Cham', 'cjm'),
    'CMR ': ('Comorian', 'swb', 'wlc', 'wni', 'zdj'),
    'COP ': ('Coptic', 'cop'),
    'COR ': ('Cornish', 'cor'),
    'COS ': ('Corsican', 'cos'),
    'CPP ': ('Creoles', 'crp', 'cpe', 'cpf', 'cpp'),
    'CRE ': ('Cree', 'cre'),
    'CRR ': ('Carrier', 'crx', 'caf'),
    'CRT ': ('Crimean Tatar', 'crh'),
    'CSB ': ('Kashubian', 'csb'),
    'CSL ': ('Church Slavonic', 'chu'),
    'CSY ': ('Czech', 'ces'),
    'CTG ': ('Chittagonian', 'ctg'),
    'CUK ': ('San Blas Kuna', 'cuk'),
    'DAN ': ('Danish', 'dan'),
    'DAR ': ('Dargwa', 'dar'),
    'DAX ': ('Dayi', 'dax'),
    'DCR ': ('Woods Cree', 'cwd'),
    'DEU ': ('German', 'deu'),
    'DGO ': ('Dogri', 'dgo'),
    'DGR ': ('Dogri', 'doi'),
    'DHG ': ('Dhangu', 'dhg'),
    'DHV ': ('Divehi (Dhivehi, Maldivian)', 'div'),
    'DIQ ': ('Dimli', 'diq'),
    'DIV ': ('Divehi (Dhivehi, Maldivian)', 'div'),
    'DJR ': ('Zarma', 'dje'),
    'DJR0': ('Djambarrpuyngu', 'djr'),
    'DNG ': ('Dangme', 'ada'),
    'DNJ ': ('Dan', 'dnj'),
    'DNK ': ('Dinka', 'din'),
    'DRI ': ('Dari', 'prs'),
    'DUJ ': ('Dhuwal', 'duj', 'dwu', 'dwy'),
    'DUN ': ('Dungan', 'dng'),
    'DZN ': ('Dzongkha', 'dzo'),
    'EBI ': ('Ebira', 'igb'),
    'ECR ': ('Eastern Cree', 'crj', 'crl'),
    'EDO ': ('Edo', 'bin'),
    'EFI ': ('Efik', 'efi'),
    'ELL ': ('Greek', 'ell'),
    'EMK ': ('Eastern Maninkakan', 'emk'),
    'ENG ': ('English', 'eng'),
    'ERZ ': ('Erzya', 'myv'),
    'ESP ': ('Spanish', 'spa'),
    'ESU ': ('Central Yupik', 'esu'),
    'ETI ': ('Estonian', 'est'),
    'EUQ ': ('Basque', 'eus'),
    'EVK ': ('Evenki', 'evn'),
    'EVN ': ('Even', 'eve'),
    'EWE ': ('Ewe', 'ewe'),
    'FAN ': ('French Antillean', 'acf'),
    'FAN0': ('Fang', 'fan'),
    'FAR ': ('Persian', 'fas'),
    'FAT ': ('Fanti', 'fat'),
    'FIN ': ('Finnish', 'fin'),
    'FJI ': ('Fijian', 'fij'),
    'FLE ': ('Dutch (Flemish)', 'vls'),
    'FMP ': ('Fe’fe’', 'fmp'),
    'FNE ': ('Forest Nenets', 'enf', 'yrk'),
    'FON ': ('Fon', 'fon'),
    'FOS ': ('Faroese', 'fao'),
    'FRA ': ('French', 'fra'),
    'FRC ': ('Cajun French', 'frc'),
    'FRI ': ('Frisian', 'fry'),
    'FRL ': ('Friulian', 'fur'),
    'FRP ': ('Arpitan', 'frp'),
    'FTA ': ('Futa', 'fuf'),
    'FUL ': ('Fulah', 'ful'),
    'FUV ': ('Nigerian Fulfulde', 'fuv'),
    'GAD ': ('Ga', 'gaa'),
    'GAE ': ('Scottish Gaelic (Gaelic)', 'gla'),
    'GAG ': ('Gagauz', 'gag'),
    'GAL ': ('Galician', 'glg'),
    'GAR ': ('Garshuni',),
    'GAW ': ('Garhwali', 'gbm'),
    'GEZ ': ('Geez', 'gez'),
    'GIH ': ('Githabul', 'gih'),
    'GIL ': ('Gilyak', 'niv'),
    'GIL0': ('Kiribati (Gilbertese)', 'gil'),
    'GKP ': ('Kpelle (Guinea)', 'gkp'),
    'GLK ': ('Gilaki', 'glk'),
    'GMZ ': ('Gumuz', 'guk'),
    'GNN ': ('Gumatj', 'gnn'),
    'GOG ': ('Gogo', 'gog'),
    'GON ': ('Gondi', 'gon'),
    'GRN ': ('Greenlandic', 'kal'),
    'GRO ': ('Garo', 'grt'),
    'GUA ': ('Guarani', 'grn'),
    'GUC ': ('Wayuu', 'guc'),
    'GUF ': ('Gupapuyngu', 'guf'),
    'GUJ ': ('Gujarati', 'guj'),
    'GUZ ': ('Gusii', 'guz'),
    'HAI ': ('Haitian (Haitian Creole)', 'hat'),
    'HAL ': ('Halam (Falam Chin)', 'flm', 'cfm', 'rnl'),
    'HAR ': ('Harauti', 'hoj'),
    'HAU ': ('Hausa', 'hau'),
    'HAW ': ('Hawaiian', 'haw'),
    'HAY ': ('Haya', 'hay'),
    'HAZ ': ('Hazaragi', 'haz'),
    'HBN ': ('Hammer-Banna', 'amf'),
    'HER ': ('Herero', 'her'),
    'HIL ': ('Hiligaynon', 'hil'),
    'HIN ': ('Hindi', 'hin'),
    'HMA ': ('High Mari', 'mrj'),
    'HMN ': ('Hmong', 'hmn'),
    'HMO ': ('Hiri Motu', 'hmo'),
    'HND ': ('Hindko', 'hno', 'hnd'),
    'HO  ': ('Ho', 'hoc'),
    'HRI ': ('Harari', 'har'),
    'HRV ': ('Croatian', 'hrv'),
    'HUN ': ('Hungarian', 'hun'),
    'HYE ': ('Armenian', 'hye', 'hyw'),
    'HYE0': ('Armenian East', 'hye'),
    'IBA ': ('Iban', 'iba'),
    'IBB ': ('Ibibio', 'ibb'),
    'IBO ': ('Igbo', 'ibo'),
    'IJO ': ('Ijo languages', 'ijc', 'ijo'),
    'IDO ': ('Ido', 'ido'),
    'ILE ': ('Interlingue', 'ile'),
    'ILO ': ('Ilokano', 'ilo'),
    'INA ': ('Interlingua', 'ina'),
    'IND ': ('Indonesian', 'ind'),
    'ING ': ('Ingush', 'inh'),
    'INU ': ('Inuktitut', 'iku'),
    'IPK ': ('Inupiat', 'ipk'),
    'IPPH': ('Phonetic transcription—IPA conventions',),
    'IRI ': ('Irish', 'gle'),
    'IRT ': ('Irish Traditional', 'gle'),
    'ISL ': ('Icelandic', 'isl'),
    'ISM ': ('Inari Sami', 'smn'),
    'ITA ': ('Italian', 'ita'),
    'IWR ': ('Hebrew', 'heb'),
    'JAM ': ('Jamaican Creole', 'jam'),
    'JAN ': ('Japanese', 'jpn'),
    'JAV ': ('Javanese', 'jav'),
    'JBO ': ('Lojban', 'jbo'),
    'JCT ': ('Krymchak', 'jct'),
    'JII ': ('Yiddish', 'yid'),
    'JUD ': ('Ladino', 'lad'),
    'JUL ': ('Jula', 'dyu'),
    'KAB ': ('Kabardian', 'kbd'),
    'KAB0': ('Kabyle', 'kab'),
    'KAC ': ('Kachchi', 'kfr'),
    'KAL ': ('Kalenjin', 'kln'),
    'KAN ': ('Kannada', 'kan'),
    'KAR ': ('Karachay', 'krc'),
    'KAT ': ('Georgian', 'kat'),
    'KAZ ': ('Kazakh', 'kaz'),
    'KDE ': ('Makonde', 'kde'),
    'KEA ': ('Kabuverdianu (Crioulo)', 'kea'),
    'KEB ': ('Kebena', 'ktb'),
    'KEK ': ('Kekchi', 'kek'),
    'KGE ': ('Khutsuri Georgian', 'kat'),
    'KHA ': ('Khakass', 'kjh'),
    'KHK ': ('Khanty-Kazim', 'kca'),
    'KHM ': ('Khmer', 'khm'),
    'KHS ': ('Khanty-Shurishkar', 'kca'),
    'KHT ': ('Khamti Shan', 'kht'),
    'KHV ': ('Khanty-Vakhi', 'kca'),
    'KHW ': ('Khowar', 'khw'),
    'KIK ': ('Kikuyu (Gikuyu)', 'kik'),
    'KIR ': ('Kirghiz (Kyrgyz)', 'kir'),
    'KIS ': ('Kisii', 'kqs', 'kss'),
    'KIU ': ('Kirmanjki', 'kiu'),
    'KJD ': ('Southern Kiwai', 'kjd'),
    'KJP ': ('Eastern Pwo Karen', 'kjp'),
    'KJZ ': ('Bumthangkha', 'kjz'),
    'KKN ': ('Kokni', 'kex'),
    'KLM ': ('Kalmyk', 'xal'),
    'KMB ': ('Kamba', 'kam'),
    'KMN ': ('Kumaoni', 'kfy'),
    'KMO ': ('Komo', 'kmw'),
    'KMS ': ('Komso', 'kxc'),
    'KMZ ': ('Khorasani Turkic', 'kmz'),
    'KNR ': ('Kanuri', 'kau'),
    'KOD ': ('Kodagu', 'kfa'),
    'KOH ': ('Korean Old Hangul', 'okm'),
    'KOK ': ('Konkani', 'kok'),
    'KON ': ('Kikongo', 'ktu'),
    'KOM ': ('Komi', 'kom'),
    'KON0': ('Kongo', 'kon'),
    'KOP ': ('Komi-Permyak', 'koi'),
    'KOR ': ('Korean', 'kor'),
    'KOS ': ('Kosraean', 'kos'),
    'KOZ ': ('Komi-Zyrian', 'kpv'),
    'KPL ': ('Kpelle', 'kpe'),
    'KRI ': ('Krio', 'kri'),
    'KRK ': ('Karakalpak', 'kaa'),
    'KRL ': ('Karelian', 'krl'),
    'KRM ': ('Karaim', 'kdr'),
    'KRN ': ('Karen', 'kar'),
    'KRT ': ('Koorete', 'kqy'),
    'KSH ': ('Kashmiri', 'kas'),
    'KSH0': ('Ripuarian', 'ksh'),
    'KSI ': ('Khasi', 'kha'),
    'KSM ': ('Kildin Sami', 'sjd'),
    'KSW ': ('S’gaw Karen', 'ksw'),
    'KUA ': ('Kuanyama', 'kua'),
    'KUI ': ('Kui', 'kxu'),
    'KUL ': ('Kulvi', 'kfx'),
    'KUM ': ('Kumyk', 'kum'),
    'KUR ': ('Kurdish', 'kur'),
    'KUU ': ('Kurukh', 'kru'),
    'KUY ': ('Kuy', 'kdt'),
    'KYK ': ('Koryak', 'kpy'),
    'KYU ': ('Western Kayah', 'kyu'),
    'LAD ': ('Ladin', 'lld'),
    'LAH ': ('Lahuli', 'bfu'),
    'LAK ': ('Lak', 'lbe'),
    'LAM ': ('Lambani', 'lmn'),
    'LAO ': ('Lao', 'lao'),
    'LAT ': ('Latin', 'lat'),
    'LAZ ': ('Laz', 'lzz'),
    'LCR ': ('L-Cree', 'crm'),
    'LDK ': ('Ladakhi', 'lbj'),
    'LEZ ': ('Lezgi', 'lez'),
    'LIJ ': ('Ligurian', 'lij'),
    'LIM ': ('Limburgish', 'lim'),
    'LIN ': ('Lingala', 'lin'),
    'LIS ': ('Lisu', 'lis'),
    'LJP ': ('Lampung', 'ljp'),
    'LKI ': ('Laki', 'lki'),
    'LMA ': ('Low Mari', 'mhr'),
    'LMB ': ('Limbu', 'lif'),
    'LMO ': ('Lombard', 'lmo'),
    'LMW ': ('Lomwe', 'ngl'),
    'LOM ': ('Loma', 'lom'),
    'LRC ': ('Luri', 'lrc', 'luz', 'bqi', 'zum'),
    'LSB ': ('Lower Sorbian', 'dsb'),
    'LSM ': ('Lule Sami', 'smj'),
    'LTH ': ('Lithuanian', 'lit'),
    'LTZ ': ('Luxembourgish', 'ltz'),
    'LUA ': ('Luba-Lulua', 'lua'),
    'LUB ': ('Luba-Katanga', 'lub'),
    'LUG ': ('Ganda', 'lug'),
    'LUH ': ('Luyia', 'luy'),
    'LUO ': ('Luo', 'luo'),
    'LVI ': ('Latvian', 'lav'),
    'MAD ': ('Madura', 'mad'),
    'MAG ': ('Magahi', 'mag'),
    'MAH ': ('Marshallese', 'mah'),
    'MAJ ': ('Majang', 'mpe'),
    'MAK ': ('Makhuwa', 'vmw'),
    'MAL ': ('Malayalam', 'mal'),
    'MAM ': ('Mam', 'mam'),
    'MAN ': ('Mansi', 'mns'),
    'MAP ': ('Mapudungun', 'arn'),
    'MAR ': ('Marathi', 'mar'),
    'MAW ': ('Marwari', 'mwr', 'dhd', 'rwr', 'mve', 'wry', 'mtr', 'swv'),
    'MBN ': ('Mbundu', 'kmb'),
    'MBO ': ('Mbo', 'mbo'),
    'MCH ': ('Manchu', 'mnc'),
    'MCR ': ('Moose Cree', 'crm'),
    'MDE ': ('Mende', 'men'),
    'MDR ': ('Mandar', 'mdr'),
    'MEN ': ('Me’en', 'mym'),
    'MER ': ('Meru', 'mer'),
    'MFA ': ('Pattani Malay', 'mfa'),
    'MFE ': ('Morisyen', 'mfe'),
    'MIN ': ('Minangkabau', 'min'),
    'MIZ ': ('Mizo', 'lus'),
    'MKD ': ('Macedonian', 'mkd'),
    'MKR ': ('Makasar', 'mak'),
    'MKW ': ('Kituba', 'mkw'),
    'MLE ': ('Male', 'mdy'),
    'MLG ': ('Malagasy', 'mlg'),
    'MLN ': ('Malinke', 'mlq'),
    'MLR ': ('Malayalam Reformed', 'mal'),
    'MLY ': ('Malay', 'msa'),
    'MND ': ('Mandinka', 'mnk'),
    'MNG ': ('Mongolian', 'mon'),
    'MNI ': ('Manipuri', 'mni'),
    'MNK ': ('Maninka', 'man', 'mnk', 'myq', 'mku', 'msc', 'emk', 'mwk', 'mlq'),
    'MNX ': ('Manx', 'glv'),
    'MOH ': ('Mohawk', 'moh'),
    'MOK ': ('Moksha', 'mdf'),
    'MOL ': ('Moldavian', 'mol'),
    'MON ': ('Mon', 'mnw'),
    'MOR ': ('Moroccan',),
    'MOS ': ('Mossi', 'mos'),
    'MRI ': ('Maori', 'mri'),
    'MTH ': ('Maithili', 'mai'),
    'MTS ': ('Maltese', 'mlt'),
    'MUN ': ('Mundari', 'unr'),
    'MUS ': ('Muscogee', 'mus'),
    'MWL ': ('Mirandese', 'mwl'),
    'MWW ': ('Hmong Daw', 'mww'),
    'MYN ': ('Mayan', 'myn'),
    'MZN ': ('Mazanderani', 'mzn'),
    'NAG ': ('Naga-Assamese', 'nag'),
    'NAH ': ('Nahuatl', 'nah'),
    'NAN ': ('Nanai', 'gld'),
    'NAP ': ('Neapolitan', 'nap'),
    'NAS ': ('Naskapi', 'nsk'),
    'NAU ': ('Nauruan', 'nau'),
    'NAV ': ('Navajo', 'nav'),
    'NCR ': ('N-Cree', 'csw'),
    'NDB ': ('Ndebele', 'nbl', 'nde'),
    'NDC ': ('Ndau', 'ndc'),
    'NDG ': ('Ndonga', 'ndo'),
    'NDS ': ('Low Saxon', 'nds'),
    'NEP ': ('Nepali', 'nep'),
    'NEW ': ('Newari', 'new'),
    'NGA ': ('Ngbaka', 'nga'),
    'NGR ': ('Nagari',),
    'NHC ': ('Norway House Cree', 'csw'),
    'NIS ': ('Nisi', 'dap', 'njz', 'tgj'),
    'NIU ': ('Niuean', 'niu'),
    'NKL ': ('Nyankole', 'nyn'),
    'NKO ': ('N’Ko', 'nqo'),
    'NLD ': ('Dutch', 'nld'),
    'NOE ': ('Nimadi', 'noe'),
    'NOG ': ('Nogai', 'nog'),
    'NOR ': ('Norwegian', 'nob'),
    'NOV ': ('Novial', 'nov'),
    'NSM ': ('Northern Sami', 'sme'),
    'NSO ': ('Sotho, Northern', 'nso'),
    'NTA ': ('Northern Tai', 'nod'),
    'NTO ': ('Esperanto', 'epo'),
    'NYM ': ('Nyamwezi', 'nym'),
    'NYN ': ('Norwegian Nynorsk (Nynorsk, Norwegian)', 'nno'),
    'NZA ': ('Mbembe Tigon', 'nza'),
    'OCI ': ('Occitan', 'oci'),
    'OCR ': ('Oji-Cree', 'ojs'),
    'OJB ': ('Ojibway', 'oji'),
    'ORI ': ('Odia (formerly Oriya)', 'ori'),
    'ORO ': ('Oromo', 'orm'),
    'OSS ': ('Ossetian', 'oss'),
    'PAA ': ('Palestinian Aramaic', 'sam'),
    'PAG ': ('Pangasinan', 'pag'),
    'PAL ': ('Pali', 'pli'),
    'PAM ': ('Pampangan', 'pam'),
    'PAN ': ('Punjabi', 'pan'),
    'PAP ': ('Palpa', 'plp'),
    'PAP0': ('Papiamentu', 'pap'),
    'PAS ': ('Pashto', 'pus'),
    'PAU ': ('Palauan', 'pau'),
    'PCC ': ('Bouyei', 'pcc'),
    'PCD ': ('Picard', 'pcd'),
    'PDC ': ('Pennsylvania German', 'pdc'),
    'PGR ': ('Polytonic Greek', 'ell'),
    'PHK ': ('Phake', 'phk'),
    'PIH ': ('Norfolk', 'pih'),
    'PIL ': ('Filipino', 'fil'),
    'PLG ': ('Palaung', 'pce', 'rbb', 'pll'),
    'PLK ': ('Polish', 'pol'),
    'PMS ': ('Piemontese', 'pms'),
    'PNB ': ('Western Panjabi', 'pnb'),
    'POH ': ('Pocomchi', 'poh'),
    'PON ': ('Pohnpeian', 'pon'),
    'PRO ': ('Provençal / Old Provençal', 'pro'),
    'PTG ': ('Portuguese', 'por'),
    'PWO ': ('Western Pwo Karen', 'pwo'),
    'QIN ': ('Chin', 'bgr', 'cnh', 'cnw', 'czt', 'sez', 'tcp', 'csy', 'ctd', 'flm', 'pck', 'tcz', 'zom', 'cmr', 'dao', 'hlt', 'cka', 'cnk', 'mrh', 'cbl', 'cnb', 'csh'),
    'QUC ': ('K’iche’', 'quc'),
    'QUH ': ('Quechua (Bolivia)', 'quh'),
    'QUZ ': ('Quechua', 'quz'),
    'QVI ': ('Quechua (Ecuador)', 'qvi'),
    'QWH ': ('Quechua (Peru)', 'qwh'),
    'RAJ ': ('Rajasthani', 'raj'),
    'RAR ': ('Rarotongan', 'rar'),
    'RBU ': ('Russian Buriat', 'bxr'),
    'RCR ': ('R-Cree', 'atj'),
    'REJ ': ('Rejang', 'rej'),
    'RIA ': ('Riang', 'ria'),
    'RIF ': ('Tarifit', 'rif'),
    'RIT ': ('Ritarungo', 'rit'),
    'RKW ': ('Arakwal', 'rkw'),
    'RMS ': ('Romansh', 'roh'),
    'RMY ': ('Vlax Romani', 'rmy'),
    'ROM ': ('Romanian', 'ron'),
    'ROY ': ('Romany', 'rom'),
    'RSY ': ('Rusyn', 'rue'),
    'RTM ': ('Rotuman', 'rtm'),
    'RUA ': ('Kinyarwanda', 'kin'),
    'RUN ': ('Rundi', 'run'),
    'RUP ': ('Aromanian', 'rup'),
    'RUS ': ('Russian', 'rus'),
    'SAD ': ('Sadri', 'sck'),
    'SAN ': ('Sanskrit', 'san'),
    'SAS ': ('Sasak', 'sas'),
    'SAT ': ('Santali', 'sat'),
    'SAY ': ('Sayisi', 'chp'),
    'SCN ': ('Sicilian', 'scn'),
    'SCO ': ('Scots', 'sco'),
    'SEK ': ('Sekota', 'xan'),
    'SEL ': ('Selkup', 'sel'),
    'SGA ': ('Old Irish', 'sga'),
    'SGO ': ('Sango', 'sag'),
    'SGS ': ('Samogitian', 'sgs'),
    'SHI ': ('Tachelhit', 'shi'),
    'SHN ': ('Shan', 'shn'),
    'SIB ': ('Sibe', 'sjo'),
    'SID ': ('Sidamo', 'sid'),
    'SIG ': ('Silte Gurage', 'xst', 'stv', 'wle'),
    'SKS ': ('Skolt Sami', 'sms'),
    'SKY ': ('Slovak', 'slk'),
    'SCS ': ('North Slavey', 'scs'),
    'SLA ': ('Slavey', 'scs', 'xsl'),
    'SLV ': ('Slovenian', 'slv'),
    'SML ': ('Somali', 'som'),
    'SMO ': ('Samoan', 'smo'),
    'SNA ': ('Sena', 'seh'),
    'SNA0': ('Shona', 'sna'),
    'SND ': ('Sindhi', 'snd'),
    'SNH ': ('Sinhala (Sinhalese)', 'sin'),
    'SNK ': ('Soninke', 'snk'),
    'SOG ': ('Sodo Gurage', 'gru'),
    'SOP ': ('Songe', 'sop'),
    'SOT ': ('Sotho, Southern', 'sot'),
    'SQI ': ('Albanian', 'sqi'),
    'SRB ': ('Serbian', 'srp'),
    'SRD ': ('Sardinian', 'srd'),
    'SRK ': ('Saraiki', 'skr'),
    'SRR ': ('Serer', 'srr'),
    'SSL ': ('South Slavey', 'xsl'),
    'SSM ': ('Southern Sami', 'sma'),
    'STQ ': ('Saterland Frisian', 'stq'),
    'SUK ': ('Sukuma', 'suk'),
    'SUN ': ('Sundanese', 'sun'),
    'SUR ': ('Suri', 'suq'),
    'SVA ': ('Svan', 'sva'),
    'SVE ': ('Swedish', 'swe'),
    'SWA ': ('Swadaya Aramaic', 'aii'),
    'SWK ': ('Swahili', 'swa'),
    'SWZ ': ('Swati', 'ssw'),
    'SXT ': ('Sutu', 'ngo'),
    'SXU ': ('Upper Saxon', 'sxu'),
    'SYL ': ('Sylheti', 'syl'),
    'SYR ': ('Syriac', 'aii', 'amw', 'cld', 'syc', 'syr', 'tru'),
    'SYRE': ('Syre', 'syc', 'syr'),
    'SYRJ': ('Syrj', 'syc', 'syr'),
    'SYRN': ('Syrn', 'syc', 'syr'),
    'SZL ': ('Silesian', 'szl'),
    'TAB ': ('Tabasaran', 'tab'),
    'TAJ ': ('Tajiki', 'tgk'),
    'TAM ': ('Tamil', 'tam'),
    'TAT ': ('Tatar', 'tat'),
    'TCR ': ('TH-Cree', 'cwd'),
    'TDD ': ('Dehong Dai', 'tdd'),
    'TEL ': ('Telugu', 'tel'),
    'TET ': ('Tetum', 'tet'),
    'TGL ': ('Tagalog', 'tgl'),
    'TGN ': ('Tongan', 'ton'),
    'TGR ': ('Tigre', 'tig'),
    'TGY ': ('Tigrinya', 'tir'),
    'THA ': ('Thai', 'tha'),
    'THT ': ('Tahitian', 'tah'),
    'TIB ': ('Tibetan', 'bod'),
    'TIV ': ('Tiv', 'tiv'),
    'TKM ': ('Turkmen', 'tuk'),
    'TMH ': ('Tamashek', 'tmh'),
    'TMN ': ('Temne', 'tem'),
    'TNA ': ('Tswana', 'tsn'),
    'TNE ': ('Tundra Nenets', 'enh', 'yrk'),
    'TNG ': ('Tonga', 'toi'),
    'TOD ': ('Todo', 'xal'),
    'TOD0': ('Toma', 'tod'),
    'TPI ': ('Tok Pisin', 'tpi'),
    'TRK ': ('Turkish', 'tur'),
    'TSG ': ('Tsonga', 'tso'),
    'TSJ ': ('Tshangla', 'tsj'),
    'TUA ': ('Turoyo Aramaic', 'tru'),
    'TUM ': ('Tulu', 'tum'),
    'TUL ': ('Tumbuka', 'tcy'),
    'TUV ': ('Tuvin', 'tyv'),
    'TVL ': ('Tuvalu', 'tvl'),
    'TWI ': ('Twi', 'aka'),
    'TYZ ': ('Tày', 'tyz'),
    'TZM ': ('Tamazight', 'tzm'),
    'TZO ': ('Tzotzil', 'tzo'),
    'UDM ': ('Udmurt', 'udm'),
    'UKR ': ('Ukrainian', 'ukr'),
    'UMB ': ('Umbundu', 'umb'),
    'URD ': ('Urdu', 'urd'),
    'USB ': ('Upper Sorbian', 'hsb'),
    'UYG ': ('Uyghur', 'uig'),
    'UZB ': ('Uzbek', 'uzb'),
    'VEC ': ('Venetian', 'vec'),
    'VEN ': ('Venda', 'ven'),
    'VIT ': ('Vietnamese', 'vie'),
    'VOL ': ('Volapük', 'vol'),
    'VRO ': ('Võro', 'vro'),
    'WA  ': ('Wa', 'wbm'),
    'WAG ': ('Wagdi', 'wbr'),
    'WAR ': ('Waray-Waray', 'war'),
    'WCR ': ('West-Cree', 'crk'),
    'WEL ': ('Welsh', 'cym'),
    'WLN ': ('Walloon', 'wln'),
    'WLF ': ('Wolof', 'wol'),
    'WTM ': ('Mewati', 'wtm'),
    'XBD ': ('Lü', 'khb'),
    'XKF ': ('Khengkha', 'xkf'),
    'XHS ': ('Xhosa', 'xho'),
    'XJB ': ('Minjangbal', 'xjb'),
    'XOG ': ('Soga', 'xog'),
    'XPE ': ('Kpelle (Liberia)', 'xpe'),
    'YAK ': ('Sakha', 'sah'),
    'YAO ': ('Yao', 'yao'),
    'YAP ': ('Yapese', 'yap'),
    'YBA ': ('Yoruba', 'yor'),
    'YCR ': ('Y-Cree', 'cre'),
    'YIC ': ('Yi Classic',),
    'YIM ': ('Yi Modern', 'iii'),
    'ZEA ': ('Zealandic', 'zea'),
    'ZGH ': ('Standard Moroccan Tamazight', 'zgh'),
    'ZHA ': ('Zhuang', 'zha'),
    'ZHH ': ('Chinese, Hong Kong SAR', 'zho'),
    'ZHP ': ('Chinese Phonetic', 'zho'),
    'ZHS ': ('Chinese Simplified', 'zho'),
    'ZHT ': ('Chinese Traditional', 'zho'),
    'ZND ': ('Zande', 'zne'),
    'ZUL ': ('Zulu', 'zul'),
    'ZZA ': ('Zazaki', 'zza'),
}



import io
import logging
from fontTools.ttLib import TTFont
from fontTools.pens.pointPen import PointToSegmentPen
import freetype
try:
    from ..mac.makePathFromOutline import makePathFromOutline
except:
    pass


class FTFont:

    @classmethod
    def fromPath(cls, path, **kwargs):
        with open(path, "rb") as f:
            fontData = f.read()
        return cls(fontData, **kwargs)

    def __init__(self, fontData, *, fontNumber=0, ttFont=None):
        if ttFont is None:
            stream = io.BytesIO(fontData)
            ttFont = TTFont(stream, fontNumber=fontNumber, lazy=True)
        self._ttFont = ttFont
        stream = io.BytesIO(fontData)
        self._ftFace = freetype.Face(stream, index=fontNumber)
        try:
            self._ftFace.set_char_size(self._ftFace.units_per_EM)
        except freetype.FT_Exception as e:
            logging.warning("FreeType error, possibly with unsupported pixel font: %s", e)

    def setVarLocation(self, varLocation):
        if "fvar" not in self._ttFont:
            return
        coordinates = []
        for axis in self._ttFont["fvar"].axes:
            coordinates.append(varLocation.get(axis.axisTag, axis.defaultValue))
        coordinates = [round(v * 0x10000) for v in coordinates]
        c_coordinates = (freetype.FT_Fixed * len(coordinates))(*coordinates)
        freetype.FT_Set_Var_Design_Coordinates(self._ftFace._FT_Face, len(coordinates), c_coordinates)

    def drawGlyphToPointPen(self, glyphName, pen):
        glyphID = self._ttFont.getGlyphID(glyphName)
        face = self._ftFace
        face.load_glyph(glyphID, freetype.FT_LOAD_NO_SCALE)
        contours = (i + 1 for i in face.glyph.outline.contours)
        points = face.glyph.outline.points
        flags = face.glyph.outline.tags
        curveType = "curve" if any(t & 0x02 for t in flags) else "qcurve"
        fromIndex = 0
        for toIndex in contours:
            cPoints = points[fromIndex:toIndex]
            cFlags = flags[fromIndex:toIndex]
            pen.beginPath()
            for i in range(len(cPoints)):
                if not cFlags[i] & 0x01:
                    segmentType = None
                elif cFlags[i - 1] & 0x01:
                    segmentType = "line"
                else:
                    segmentType = curveType
                pen.addPoint(cPoints[i], segmentType)
            pen.endPath()
            fromIndex = toIndex

    def drawGlyphToPen(self, glyphName, pen):
        self.drawGlyphToPointPen(glyphName, PointToSegmentPen(pen))

    def getOutlinePath(self, glyphName):
        glyphID = self._ttFont.getGlyphID(glyphName)
        face = self._ftFace
        face.load_glyph(glyphID, freetype.FT_LOAD_NO_SCALE)
        return makePathFromOutline(face.glyph.outline._FT_Outline)



import unittest
from coldtype.time.nle.ascii import AsciiTimeline

at = AsciiTimeline(1, """
                                    |
[a          ]
        [b          ]
                [c          ]
     ]                  [d          
""")

class TestTime(unittest.TestCase):
    def test_ascii_timeline(self):
        self.assertEqual(at.duration, 36)
        self.assertEqual(at[0].start, 0)
        self.assertEqual(at[1].end, 21)
        self.assertEqual(at[2].start, 16)
        self.assertEqual(at[-1].end, 41)
        self.assertEqual(len(at.clips), 4)

if __name__ == "__main__":
    unittest.main()


import enum
from coldtype.time.timeline import Timeline, Timeable


class AsciiTimeline(Timeline):
    __name__ = "AsciiTimeline"

    def __init__(self,
        multiplier:int,
        fps:float,
        ascii:str=None,
        sort=False,
        **kwargs
        ):
        if isinstance(fps, str):
            ascii = fps
            fps = 30
        lines = [l.rstrip() for l in ascii.splitlines() if l.strip()]
        ml = max([len(l) for l in lines]) - 1
        super().__init__(multiplier*ml, fps=fps, **kwargs)

        self.multiplier = multiplier
        
        clips = []

        unclosed_clip = None
        for lidx, l in enumerate(lines):
            clip_start = None
            clip_name = None
            if unclosed_clip:
                clip_start, clip_name = unclosed_clip
                unclosed_clip = None
            looped_clip_end = None
            for idx, c in enumerate(l):
                if c == "]":
                    if clip_start is not None and clip_name is not None:
                        clips.append(Timeable(
                            clip_start,
                            (idx+1)*multiplier,
                            name=clip_name,
                            data=dict(line=lidx),
                            timeline=self))
                    else:
                        looped_clip_end = idx*multiplier
                    clip_start = None
                    clip_name = None
                elif c == "[":
                    clip_start = idx*multiplier
                    clip_name = ""
                elif c not in [" ", "-", "|", "<", ">"]:
                    clip_name += c
            
            if looped_clip_end:
                if clip_start is not None and clip_name is not None:
                    clips.append(Timeable(
                        clip_start,
                        self.duration+looped_clip_end,
                        name=clip_name,
                        data=dict(line=idx),
                        timeline=self))
                    clip_start = None
                    clip_name = None
            
            if clip_start is not None and clip_name is not None:
                unclosed_clip = (clip_start, clip_name)
        
        if sort:
            self.clips = sorted(clips, key=lambda c: c.name)
        else:
            self.clips = clips
        
        for cidx, clip in enumerate(self.clips):
            clip.index = cidx

    def __getitem__(self, item):
        if isinstance(item, str):
            for c in self.clips:
                if c.name == item:
                    return c
        else:
            return self.clips[item]
    
    def now(self, fi, line=None, first=False, filter_fn=None):
        matches = []
        for clip in self.clips:
            if clip.start <= fi < clip.end:
                if line is not None:
                    if clip.data["line"] != line:
                        continue
                matches.append(clip)
        if filter_fn:
            matches = list(filter(filter_fn, matches))
        if first:
            return matches[0] if matches else None
        return matches


from collections import defaultdict
import functools
import io
import itertools
from fontTools.ttLib import TTFont
import uharfbuzz as hb


class GlyphInfo:

    def __init__(self, gid, name, cluster, dx, dy, ax, ay):
        self.gid = gid
        self.name = name
        self.cluster = cluster
        self.dx = dx
        self.dy = dy
        self.ax = ax
        self.ay = ay

    def __repr__(self):
        args = (f"{a}={repr(getattr(self, a))}"
                for a in ["gid", "name", "cluster", "dx", "dy", "ax", "ay"])
        return f"{self.__class__.__name__}({', '.join(args)})"


#
# To generalize, HB needs
# - a callback to map a character to a glyph ID
# - a callback to get the advance width for a glyph ID
# - a callback to get the advance height for a glyph ID
# - either data for a minimal ttf, or a callback getting individual table data
#   (the latter is broken in current uharfbuzz)
# To make our shaper work, we need
# - to provide a glyph order so we can map glyph IDs to glyph names
# - apart from providing a glyph order, we want our callbacks to deal
#   with glyph names, not glyph IDs.
#


def _getGlyphIDFunc(font, char, shaper):
    glyphName = shaper.getGlyphNameFromCodePoint(char)
    if glyphName is None:
        return 0  # .notdef
    glyphID = shaper.getGlyphID(glyphName, 0)
    return glyphID


def _getHorizontalAdvanceFunc(font, glyphID, shaper):
    glyphName = shaper.glyphOrder[glyphID]
    return shaper.getHorizontalAdvance(glyphName)


def _getVerticalAdvanceFunc(font, glyphID, shaper):
    glyphName = shaper.glyphOrder[glyphID]
    return shaper.getVerticalAdvance(glyphName)


def _getVerticalOriginFunc(font, glyphID, shaper):
    glyphName = shaper.glyphOrder[glyphID]
    return shaper.getVerticalOrigin(glyphName)


_stylisticSets = {f"ss{i:02}" for i in range(1, 21)}


class HBShape:

    @classmethod
    def fromPath(cls, path, **kwargs):
        with open(path, "rb") as f:
            fontData = f.read()
        return cls(fontData, **kwargs)

    def __init__(self, fontData, *, fontNumber=0,
                 getGlyphNameFromCodePoint=None,
                 getHorizontalAdvance=None,
                 getVerticalAdvance=None,
                 getVerticalOrigin=None,
                 ttFont=None):
        self._fontData = fontData
        self._fontNumber = fontNumber
        self.face = hb.Face(fontData, fontNumber)
        self.font = hb.Font(self.face)

        if ttFont is None:
            f = io.BytesIO(self._fontData)
            ttFont = TTFont(f, fontNumber=self._fontNumber, lazy=True)
        self._ttFont = ttFont
        self.glyphOrder = ttFont.getGlyphOrder()

        if getGlyphNameFromCodePoint is None and getHorizontalAdvance is not None:
            def _getGlyphNameFromCodePoint(cmap, codePoint):
                return cmap.get(codePoint)
            getGlyphNameFromCodePoint = functools.partial(_getGlyphNameFromCodePoint, self._ttFont.getBestCmap())

        if getGlyphNameFromCodePoint is not None:
            assert getHorizontalAdvance is not None

        self.getGlyphNameFromCodePoint = getGlyphNameFromCodePoint
        self.getHorizontalAdvance = getHorizontalAdvance
        self.getVerticalAdvance = getVerticalAdvance
        self.getVerticalOrigin = getVerticalOrigin

        if getGlyphNameFromCodePoint is not None and getHorizontalAdvance is not None:
            self._funcs = hb.FontFuncs.create()
            self._funcs.set_nominal_glyph_func(_getGlyphIDFunc, self)
            self._funcs.set_glyph_h_advance_func(_getHorizontalAdvanceFunc, self)
            if getVerticalAdvance is not None:
                self._funcs.set_glyph_v_advance_func(_getVerticalAdvanceFunc, self)
            if getVerticalOrigin is not None:
                self._funcs.set_glyph_v_origin_func(_getVerticalOriginFunc, self)
        else:
            self._funcs = None

    def getFeatures(self, otTableTag):
        features = set()
        for scriptIndex, script in enumerate(hb.ot_layout_table_get_script_tags(self.face, otTableTag)):
            langIdices = list(range(len(hb.ot_layout_script_get_language_tags(self.face, otTableTag, scriptIndex))))
            langIdices.append(0xFFFF)
            for langIndex in langIdices:
                features.update(hb.ot_layout_language_get_feature_tags(self.face, otTableTag, scriptIndex, langIndex))
        return features

    def getStylisticSetNames(self):
        tags = _stylisticSets & set(self.getFeatures("GSUB"))
        if not tags:
            return {}
        gsubTable = self._ttFont.get("GSUB")
        nameTable = self._ttFont.get("name")
        if gsubTable is None or nameTable is None:
            return {}
        gsubTable = gsubTable.table
        names = {}
        for feature in gsubTable.FeatureList.FeatureRecord:
            tag = feature.FeatureTag
            if tag in tags and tag not in names:
                feaParams = feature.Feature.FeatureParams
                if feaParams is not None:
                    nameRecord = nameTable.getName(feaParams.UINameID, 3, 1)
                    if nameRecord is not None:
                        names[tag] = nameRecord.toUnicode()
        return names

    def getScriptsAndLanguages(self, otTableTag):
        scriptsAndLanguages = {}
        for scriptIndex, script in enumerate(hb.ot_layout_table_get_script_tags(self.face, otTableTag)):
            scriptsAndLanguages[script] = set(hb.ot_layout_script_get_language_tags(self.face, otTableTag, scriptIndex))
        return scriptsAndLanguages

    def getGlyphID(self, glyphName, default=0):
        try:
            return self._ttFont.getGlyphID(glyphName)
        except KeyError:
            return default

    def shape(self, text, *, features=None, varLocation=None,
              direction=None, language=None, script=None):
        if features is None:
            features = {}
        if varLocation is None:
            varLocation = {}

        self.font.scale = (self.face.upem, self.face.upem)
        self.font.set_variations(varLocation)

        hb.ot_font_set_funcs(self.font)

        if self._funcs is not None:
            self.font.funcs = self._funcs

        buf = hb.Buffer.create()
        buf.add_str(str(text))  # add_str() does not accept str subclasses
        buf.guess_segment_properties()

        buf.cluster_level = hb.BufferClusterLevel.MONOTONE_CHARACTERS

        if direction is not None:
            buf.direction = direction
        if language is not None:
            buf.set_language_from_ot_tag(language)
        if script is not None:
            buf.set_script_from_ot_tag(script)

        hb.shape(self.font, buf, features)

        glyphOrder = self.glyphOrder
        infos = []
        for info, pos in zip(buf.glyph_infos, buf.glyph_positions):
            infos.append(GlyphInfo(info.codepoint, glyphOrder[info.codepoint], info.cluster, *pos.position))

        return infos


def characterGlyphMapping(clusters, numChars):
    """This implements character to glyph mapping and vice versa, using
    cluster information from HarfBuzz. It should be correct for HB
    clustering support levels 0 and 1, see:

        https://harfbuzz.github.io/working-with-harfbuzz-clusters.html

    "Each character belongs to the cluster that has the highest cluster
    value not larger than its initial cluster value.""
    """

    if clusters:
        if clusters[-1] != 0:
            assert clusters[0] == 0

    clusterToChars = {}
    charToCluster = {}
    for cl, clNext in _pairs(sorted(set(clusters)), numChars):
        chars = list(range(cl, clNext))
        clusterToChars[cl] = chars
        for char in chars:
            charToCluster[char] = cl

    glyphToChars = [clusterToChars[cl] for cl in clusters]

    charToGlyphs = defaultdict(list)
    for glyphIndex, charIndices in enumerate(glyphToChars):
        for ci in charIndices:
            charToGlyphs[ci].append(glyphIndex)

    # assert sorted(charToGlyphs) == list(range(numChars)), charToGlyphs
    charToGlyphs = [charToGlyphs[ci] for ci in sorted(charToGlyphs)]

    return glyphToChars, charToGlyphs


def _pairs(seq, sentinel):
    it = itertools.chain(seq, (sentinel,))
    prev = next(it)
    for i in it:
        yield prev, i
        prev = i



import unittest
from coldtype.grid import Grid
from coldtype.geometry import *
from coldtype.sh import sh, SHContext, SHLookup

def bounds(self):
    return "BOUNDS"

SHContext.bounds = bounds

class TestSh(unittest.TestCase):
    
    def test_sh_context(self):
        ctx = SHContext()
        ctx.locals["cool"] = "io"
        ctx.subs["□"] = lambda c: "ctx.constants.bx" if hasattr(c, "guides") and hasattr(c.guides, "bx") else "ctx.bounds()"
        ctx.subs["■"] = "_dps.bounds()"

        ctx.context_record("$", "constants", None,
            hello="'world'",
            yoyo=lambda _: "ma",
            aƒbƒc="cool $hello □")
        
        self.assertEqual(len(ctx.constants), 5)
        self.assertEqual(ctx.constants["yoyo"], "ma")
        self.assertEqual(ctx.constants["b"], "world")
        
        res = ctx.sh("cool $hello □")
        self.assertEqual(res, ["io", "world", "BOUNDS"])
    
    def test_sh_grid_arg(self):
        g = Grid(Rect(100, 100), "10 a 10", "a", "x y z")
        self.assertEqual(list(g.keyed.keys()), ["x", "y", "z"])
        
        ctx = SHContext()
        ctx.context_record("&", "guides", None, g)
        self.assertEqual(ctx.guides.x.w, 10)
        self.assertEqual(g.keyed["x"].h, 100)
        self.assertEqual(ctx.sh("&y")[0].w, 80)
        self.assertEqual(ctx.sh("&y•")[0], Point(50, 50))
        self.assertEqual(ctx.sh("&z⊢")[0], Line(Point(90, 0), Point(90, 100)))
    
    def test_sh_string_literal(self):
        ctx = SHContext()
        ctx.context_record("$", "defs", None,
            hello="'world'")

if __name__ == "__main__":
    unittest.main()


#coding=utf-8
from __future__ import division
from math import pi, radians, degrees, tan, hypot, atan2, cos, sin

"""
Custom Robofab pens — Loïc Sander, june 2015.

The MIT License (MIT)

Copyright (c) 2014 Loïc Sander

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from fontTools.pens.pointPen import PointToSegmentPen, SegmentToPointPen, ReverseContourPointPen
from fontTools.misc.bezierTools import splitCubicAtT
from fontTools.pens.basePen import BasePen

_ANGLE_EPSILON = pi/36


def calcVector(point1, point2):
    x1, y1 = point1
    x2, y2 = point2
    dx = x2 - x1
    dy = y2 - y1
    return dx, dy


def calcAngle(point1, point2):
    dx, dy = calcVector(point1, point2)
    return atan2(dy, dx)


def polarCoord(xy, angle, distance):
    x, y = xy
    nx = x + (distance * cos(angle))
    ny = y + (distance * sin(angle))
    return nx, ny


def calcArea(points):
    l = len(points)
    area = 0
    for i in range(l):
        x1, y1 = points[i]
        x2, y2 = points[(i+1)%l]
        area += (x1*y2)-(x2*y1)
    return area / 2


def firstDerivative(xy1, cxy1, cxy2, xy2, value):
    x1, y1 = xy1
    cx1, cy1 = cxy1
    cx2, cy2 = cxy2
    x2, y2 = xy2
    mx = bezierTangent(x1, cx1, cx2, x2, value)
    my = bezierTangent(y1, cy1, cy2, y2, value)
    return mx, my


def bezierTangent(a, b, c, d, t):
    # Implementation of http://stackoverflow.com/questions/4089443/find-the-tangent-of-a-point-on-a-cubic-bezier-curve-on-an-iphone
    return (-3*(1-t)**2 * a) + (3*(1-t)**2 * b) - (6*t*(1-t) * b) - (3*t**2 * c) + (6*t*(1-t) * c) + (3*t**2 * d)



class TranslationPen(BasePen):
    """
    Draw an outline resulting from the reunion of an initial contour and a translated version thereof.
    Translation is defined by an angle and a width/length.
    This kind of drawing basically produces a calligraphic effect (in a translated manner as Gerrit Noordzij puts it),
    it can also serve as a way of extruding a shape for 3D shadow effects.
    """

    def __init__(self, otherPen, frontAngle=0, frontWidth=20):
        self.otherPen = otherPen
        self.frontAngle = radians(frontAngle)
        self.offset = polarCoord((0, 0), radians(frontAngle), frontWidth)
        self.points = []


    def _moveTo(self, pt):
        self.points.append((pt, 'move'))


    def _lineTo(self, pt1):
        pt0, previousType = self.points[-1]
        angle = calcAngle(pt0, pt1)

        self.translatedLineSegment(pt0, pt1)

        self.points.append((pt1, 'line'))


    def _curveToOne(self, c1, c2, pt1):
        pt0, previousType = self.points[-1]

        newSegments = self.splitAtAngledExtremas(pt0, c1, c2, pt1)

        if len(newSegments):
            for segment in newSegments:
                pt0, c1, c2, pt1 = segment
                self.translatedCurveSegment(pt0, c1, c2, pt1)
        else:
            self.translatedCurveSegment(pt0, c1, c2, pt1)

        self.points.append((c1, None))
        self.points.append((c2, None))
        self.points.append((pt1, 'curve'))


    def endPath(self):
        self.points = []


    def closePath(self):
        previousPoint, previousType = self.points[-1]

        if previousType in ['line','curve']:
            pt0, pt1 = self.points[-1][0], self.points[0][0]
            self.translatedLineSegment(pt0, pt1)

        self.points = []


    def splitAtAngledExtremas(self, pt0, pt1, pt2, pt3):
        frontAngle = self.frontAngle
        segments = []
        for i in range(101):
            t = i / 100
            nx, ny = firstDerivative(pt0, pt1, pt2, pt3, t)
            tanAngle = atan2(ny, nx)
            if tan(frontAngle - _ANGLE_EPSILON) < tan(tanAngle) < tan(frontAngle + _ANGLE_EPSILON):
                newSegments = splitCubicAtT(pt0, pt1, pt2, pt3, t)
                if len(newSegments) > 1:
                    segments = newSegments
                    break
        return segments


    def translatedCurveSegment(self, pt0, c1, c2, pt1):
        ox, oy = self.offset
        x0, y0 = pt0
        xc1, yc1 = c1
        xc2, yc2 = c2
        x1, y1 = pt1
        pen = self.getPen([(x0, y0), (x1, y1), (x1+ox, y1+oy), (x0+ox, y0+oy)])
        pen.moveTo((x0, y0))
        pen.curveTo((xc1, yc1), (xc2, yc2), (x1, y1))
        pen.lineTo((x1+ox, y1+oy))
        pen.curveTo((xc2+ox, yc2+oy), (xc1+ox, yc1+oy), (x0+ox, y0+oy))
        pen.closePath()


    def translatedLineSegment(self, pt0, pt1):
        ox, oy = self.offset
        x0, y0 = pt0
        x1, y1 = pt1
        pen = self.getPen([(x0, y0), (x1, y1), (x1+ox, y1+oy), (x0+ox, y0+oy)])
        pen.moveTo((x0, y0))
        pen.lineTo((x1, y1))
        pen.lineTo((x1+ox, y1+oy))
        pen.lineTo((x0+ox, y0+oy))
        pen.closePath()


    def getPen(self, points):
        area = calcArea(points)
        if area < 0:
            pen = self.getReversePen()
        else:
            pen = self.otherPen
        return pen


    def getReversePen(self):
        adapterPen = PointToSegmentPen(self.otherPen)
        reversePen = ReverseContourPointPen(adapterPen)
        return SegmentToPointPen(reversePen)


    def addComponent(self, baseGlyphName, transformation):
        self.otherPen.addComponent(baseGlyphName, transformation)


import noise


def warp_fn(xa=0, ya=-1, xs=300, ys=300, speed=5, base=0, octaves=1, mult=50, rz=1024):
    if ya == -1:
        ya = xa
    def warp(x, y):
        _x = (x+xa)/xs
        _y = (y+ya)/ys
        pn = noise.pnoise3(_x, _y, speed, octaves=octaves, base=base, repeatz=rz)
        return x+pn*mult, y+pn*mult
    return warp


import time, math

from fontTools.pens.basePen import BasePen
from fontTools.pens.transformPen import TransformPen
from coldtype.geometry import Rect

try:
    from pyaxidraw import axidraw
except:
    print("Couldn’t import pyaxidraw")
    pass

MOVE_DELAY = 2

class AxiDrawPen(BasePen):
    def __init__(self, dat, page):
        super().__init__(None)
        self.dat = dat
        self.page = page
        self.ad = None
        #dat.replay(self)
        self.last_moveTo = None
    
    def _moveTo(self, p):
        self.last_moveTo = p
        time.sleep(MOVE_DELAY)
        self.ad.moveto(*p)
        time.sleep(MOVE_DELAY)

    def _lineTo(self, p):
        self.ad.lineto(*p)

    def _curveToOne(self, p1, p2, p3):
        print("CANNOT CURVE")

    def _qCurveToOne(self, p1, p2):
        print("CANNOT CURVE")

    def _closePath(self):
        # can this work?
        if self.last_moveTo:
            self.ad.lineto(*self.last_moveTo)

    def draw(self, scale=0.01, cm=False):
        self.dat.scale(scale, point=False)
        self.page = self.page.scale(scale)
        b = self.dat.bounds()
        limits = Rect(0, 0, 11, 8.5)
        if cm:
            limits = limits.scale(2.54)
        if b.x >= 0 and b.y >= 0 and b.w <= limits.w and b.h <= limits.h:
            print("Drawing!")
        else:
            print("Too big!", b)
            return False
        
        ad = axidraw.AxiDraw()
        ad.interactive()
        ad.options.units = 1 if cm else 0
        ad.options.speed_pendown = 10
        ad.options.speed_penup = 50
        ad.options.pen_rate_raise = 10

        ad.connect()
        ad.penup()
        self.ad = ad
        tp = TransformPen(self, (1, 0, 0, -1, 0, self.page.h))
        self.dat.replay(tp)
        time.sleep(MOVE_DELAY)
        ad.penup()
        ad.moveto(0,0)
        ad.disconnect()


import unittest
from coldtype.geometry import *
from coldtype.sh import sh

class TestTest(unittest.TestCase):
    def test_test(self):
        r = Rect(0, 0, 100, 100)
        self.assertEqual(sh(f"{r}TX=50"), [Rect([25.0,0,50,100])])
        self.assertEqual(sh("↗|65|10|↑"), [["NE", 65, 10, "N"]])

if __name__ == "__main__":
    unittest.main()


import math
from re import L
from typing import Callable, Optional

from fontTools.misc.transform import Transform

from coldtype.sh import sh
from coldtype.geometry import Geometrical, Rect
from coldtype.pens.draftingpen import DraftingPen
from coldtype.beziers import CurveCutter


class DraftingPens(DraftingPen):
    def __init__(self, pens=None):
        self._pens = []
        super().__init__()

        self.single_pen_class = DraftingPen
        self._in_progress_pen = None

        self.typographic = True
        self.layered = False
        self.data = {}

        self._alpha = 1
        self._parent = None
        self.container = None
        self._visible = True

        self.subs = {
            "□": "ctx.bounds()",
            "■": "_dps.bounds()"
        }

        if pens:
            for p in pens:
                self.append(p)
    
    def tree(self, out=None, depth=0) -> str:
        """Hierarchical string representation"""
        if out is None:
            out = []
        out.append(" |"*depth + " " + str(self))
        for pen in self._pens:
            if hasattr(pen, "_pens"):
                pen.tree(out=out, depth=depth+1)
            else:
                out.append(" |"*(depth+1) + " " + str(pen))
        return "\n".join(out)
    
    def __repr__(self):
        return f"<{type(self).__name__}:{len(self._pens)}>"
    
    def __len__(self):
        return len(self._pens)
    
    def __getitem__(self, index):
        return self._pens[index]
        
    def __setitem__(self, index, pen):
        self._pens[index] = pen
    
    def __iadd__(self, item):
        return self.append(item)
    
    def __add__(self, item):
        return self.append(item)
    
    def append(self, pen, allow_blank=False):
        if isinstance(pen, Geometrical):
            return self._pens.append(self.single_pen_class(pen))
        elif isinstance(pen, DraftingPen):
            self._pens.append(pen)
        else:
            try:
                for p in pen:
                    if p:
                        self._pens.append(p)
            except TypeError:
                self._pens.append(pen)
        return self
    
    def insert(self, index, pen):
        if pen:
            self._pens.insert(index, pen)
        return self
    
    def extend(self, pens):
        if hasattr(pens, "_pens"):
            self.append(pens)
        else:
            for p in pens:
                if p:
                    if hasattr(p, "value"):
                        self.append(p)
                    else:
                        self.extend(p)
        return self
    
    def reversePens(self):
        """Reverse the order of the pens; useful for overlapping glyphs from the left-to-right rather than right-to-left (as is common in OpenType applications)"""
        self._pens = list(reversed(self._pens))
        return self
    
    # Sizing
    
    def ambit(self, th=False, tv=False):
        """Get the calculated rect boundary of the DraftingPens;
        `th` means `(t)rue (h)orizontal`;
        `ty` means `(t)rue (v)ertical`;
        passing either ignores a non-bounds-derived frame
        in either dimension"""
        if self._frame and (th == False and tv == False):
            return self._frame
        elif self._frame and (th or tv) and not (th and tv):
            f = self._frame
            b = self.bounds()
            if th and tv:
                return b
            elif th:
                return Rect(b.x, f.y, b.w, f.h)
            else:
                return Rect(f.x, b.y, f.w, b.h)
        else:
            try:
                union = self._pens[0].ambit(th=th, tv=tv)
                for p in self._pens[1:]:
                    if hasattr(p, "_pens") or len(p.value) > 0:
                        union = union.union(p.ambit(th=th, tv=tv))
                return union
            except Exception as e:
                return Rect(0,0,0,0)
    
    def bounds(self):
        """Calculated bounds of a DraftingPens"""
        return self.ambit(th=1, tv=1)
    
    def gs(self, s, fn=None, tag=None, writer=None):
        return self.append(
            self.single_pen_class().gs(s,
                tag=tag, fn=fn, writer=writer,
                macros=self.macros,
                ctx=self, dps=type(self)()))
    
    def gss(self, s):
        dps = type(self)()
        xs = sh(s, ctx=self, dps=dps)
        return self.extend(dps._pens)
    
    def gsgroup(self, fn):
        grouper = self.copy(with_data=False)
        grouper.define(**self.defs.values)
        grouper.macros = self.macros
        grouper._pens = []
        res = fn(grouper)
        self.ap(res)
        return self
    
    # RecordingPen contract

    def moveTo(self, p0):
        self._in_progress_pen = self.single_pen_class()
        self._in_progress_pen.moveTo(p0)
        return self
    
    def lineTo(self, p1):
        self._in_progress_pen.lineTo(p1)
        return self
    
    def qCurveTo(self, *points):
        self._in_progress_pen.qCurveTo(*points)
        return self
    
    def curveTo(self, *points):
        self._in_progress_pen.curveTo(*points)
        return self
    
    def closePath(self):
        self._in_progress_pen.closePath()
        self.append(self._in_progress_pen)
        self._in_progress_pen = None
        return self
    
    def endPath(self):
        self._in_progress_pen.endPath()
        self.append(self._in_progress_pen)
        self._in_progress_pen = None
        return self
    
    def replay(self, pen):
        self.pen().replay(pen)
    
    def record(self, pen):
        """Alias for append"""
        if callable(pen):
            return self.append(pen(self))
        else:
            return self.append(pen)
    
    def explode(self):
        """Noop on a set"""
        return self
    
    # Drawing
    
    def pen(self):
        """A flat representation of this set as a single pen"""
        dp = self.single_pen_class()
        fps = self.collapse()
        for p in fps._pens:
            dp.record(p)
        if len(fps._pens) > 0:
            for k, attrs in fps._pens[0].attrs.items():
                dp.attr(tag=k, **attrs)
        dp.frame(self.ambit())
        return dp
    
    def collapse(self, levels=100, onself=False):
        """AKA `flatten` in some programming contexts, though
        `flatten` is a totally different function here that flattens
        outlines; this function flattens nested collections into
        one-dimensional collections"""
        pens = []
        for idx, p in enumerate(self._pens):
            if hasattr(p, "_pens") and levels > 0:
                pens.extend(p.collapse(levels=levels-1)._pens)
            else:
                pens.append(p)
        dps = self.multi_pen_class(self)(pens)
        if self.layered:
            dps.layered = True
        if onself:
            self._pens = dps._pens
            return self
        else:
            return dps
        
    def copy(self, with_data=False):
        """Get a completely new copy of this whole set of pens,
        usually done so you can duplicate and further modify a
        DATPens without mutating the original"""
        dps = type(self)()
        for p in self._pens:
            dps.append(p.copy(with_data=with_data))
        return dps
    
    def remove_blanks(self):
        """Remove blank pens"""
        nonblank_pens = []
        for pen in self._pens:
            if hasattr(pen, "_pens"):
                pen.remove_blanks()
                nonblank_pens.append(pen)
            elif len(pen.value) > 0:
                nonblank_pens.append(pen)
            #rb = pen.remove_blanks()
            #print("RB RES", rb, bool(rb))
            #if not rb:
            #    nonblank_pens.append(pen)
        self._pens = nonblank_pens
        return self
    
    def remove_overlap(self):
        for p in self._pens:
            p.removeOverlap()
        return self
    
    removeOverlap = remove_overlap
    
    def transform(self, transform, transformFrame=True):
        for p in self._pens:
            p.transform(transform, transformFrame=transformFrame)
        if transformFrame and self._frame:
            self._frame = self._frame.transform(transform)
        return self
    
    def attr(self, key="default", field=None, **kwargs):
        if field: # getting, not setting, kind of weird to return the first value?
            if len(self._pens) > 0:
                return self._pens[0].attr(key=key, field=field)
            else:
                return None
        for p in self._pens:
            p.attr(key, **kwargs)
        return self
    
    def lattr(self, tag, fn: Callable[[DraftingPen], Optional[DraftingPen]]):
        for p in self._pens:
            p.lattr(tag, fn)
        return self
    
    def round_to(self, rounding):
        """Round all values for all pens in this set to nearest multiple of rounding value (rather than places, as in `round`)"""
        for p in self._pens:
            p.round_to(rounding)
        return self
    
    def xa(self, x="centerx"):
        for pen in self:
            pen.xAlignToFrame(x)
        return self
    
    def distribute(self, v=False, tracks=None):
        off = 0
        for idx, p in enumerate(self):
            if tracks is not None and idx > 0:
                t = tracks[idx-1]
                #print(t)
                off += t
            frame = p.ambit()
            if v:
                if frame.y < 0:
                    p.translate(0, -frame.y)
                p.translate(0, off)
                off += frame.h
            else:
                if frame.x < 0:
                    p.translate(-frame.x, 0)
                p.translate(off, 0)
                off += frame.w
        return self
    
    def track(self, t, v=False):
        """Track-out/distribute elements"""
        for idx, p in enumerate(self._pens):
            frame = p.ambit()
            if v:
                p.translate(0, t*idx)
            else:
                p.translate(t*idx, 0)
        return self
    
    def track_with_width(self, t):
        """Track-out/distribute elements"""
        x = 0
        for idx, p in enumerate(self._pens):
            frame = p.ambit()
            p.translate(x + t, 0)
            x += frame.w
        return self
    
    def track_to_rect(self, rect, pullToEdges=False, r=0):
        """Distribute pens evenly within a frame"""
        if len(self) == 1:
            return self.align(rect)
        total_width = 0
        pens = self._pens
        if r:
            pens = list(reversed(pens))
        start_x = pens[0].ambit(th=pullToEdges).x
        end_x = pens[-1].ambit(th=pullToEdges).point("SE").x
        # TODO easy to knock out apostrophes here based on a callback, last "actual" frame
        total_width = end_x - start_x
        leftover_w = rect.w - total_width
        tracking_value = leftover_w / (len(self)-1)
        if pullToEdges:
            xoffset = rect.x - pens[0].bounds().x
        else:
            xoffset = rect.x - pens[0].ambit().x
        for idx, p in enumerate(pens):
            if idx == 0:
                p.translate(xoffset, 0)
            else:
                p.translate(xoffset+tracking_value*idx, 0)
        return self
    
    def interleave(self, style_fn, direction=-1, recursive=True):
        """Provide a callback-lambda to interleave new DATPens between the existing ones; useful for stroke-ing glyphs, since the stroked glyphs can be placed behind the primary filled glyphs."""
        pens = []
        for idx, p in enumerate(self._pens):
            if recursive and hasattr(p, "_pens"):
                _p = p.interleave(style_fn, direction=direction, recursive=True)
                pens.append(_p)
            else:
                try:
                    np = style_fn(idx, p.copy())
                except TypeError:
                    np = style_fn(p.copy())
                if isinstance(np, self.single_pen_class):
                    np = [np]
                if direction < 0:
                    pens.extend(np)
                pens.append(p)
                if direction > 0:
                    pens.extend(np)

        self._pens = pens
        return self
    
    def understroke(self, s=0, sw=5, outline=False, dofill=0):
        if sw == 0:
            return self
        if not outline:
            return self.interleave(lambda idx, p: p.f(s).s(s).sw(sw))
        else:
            def mod(idx, p):
                if dofill:
                    pf = p.copy()
                p.f(s).outline(sw*2)
                if dofill:
                    p.reverse().record(pf)
                return p
            return self.interleave(mod)
    
    def distribute_on_path(self, path, offset=0, cc=None, notfound=None, center=False):
        if cc:
            cutter = cc
        else:
            cutter = CurveCutter(path)
        if center is not False:
            offset = (cutter.length-self.bounds().w)/2 + center
        limit = len(self._pens)
        for idx, p in enumerate(self._pens):
            f = p.ambit()
            bs = f.y
            ow = offset + f.x + f.w / 2
            #if ow < 0:
            #    if notfound:
            #        notfound(p)
            if ow > cutter.length:
                limit = min(idx, limit)
            else:
                _p, tangent = cutter.subsegmentPoint(end=ow)
                x_shift = bs * math.cos(math.radians(tangent))
                y_shift = bs * math.sin(math.radians(tangent))
                t = Transform()
                t = t.translate(_p[0] + x_shift - f.x, _p[1] + y_shift - f.y)
                t = t.translate(f.x, f.y)
                t = t.rotate(math.radians(tangent-90))
                t = t.translate(-f.x, -f.y)
                t = t.translate(-f.w*0.5)
                p.transform(t)

        if limit < len(self._pens):
            self._pens = self._pens[0:limit]
        return self
    
    # deprecated
    distributeOnPath = distribute_on_path

    def map(self, fn: Callable[[int, DraftingPen], Optional[DraftingPen]]):
        """Apply `fn` to all top-level pen(s) in this set;
        if `fn` returns a value, it will overwrite
        the pen it was given as an argument;
        fn lambda receives `idx, p` as arguments"""
        for idx, p in enumerate(self._pens):
            result = fn(idx, p)
            if result:
                self._pens[idx] = result
        return self
    
    def mmap(self, fn: Callable[[int, DraftingPen], None]):
        """Apply `fn` to all top-level pen(s) in this set but
        do not look at return value; first m in mmap
        stands for `mutate`;
        fn lambda receives `idx, p` as arguments"""
        for idx, p in enumerate(self._pens):
            fn(idx, p)
        return self
    
    def filter(self, fn: Callable[[int, DraftingPen], bool]):
        """Filter top-level pen(s)"""
        dps = self.multi_pen_class()
        for idx, p in enumerate(self._pens):
            if fn(idx, p):
                dps.append(p)
        #self._pens = dps._pens
        #return self
        return dps
    
    def pmap(self, fn):
        """Apply `fn` to all individal pens, recursively"""
        for idx, p in enumerate(self._pens):
            if hasattr(p, "_pens"):
                p.pmap(fn)
            else:
                fn(idx, p)
        return self
    
    def pfilter(self, fn):
        """Filter all pens, recursively"""
        to_keep = []
        for idx, p in enumerate(self._pens):
            if hasattr(p, "_pens"):
                matches = p.pfilter(fn)
                if len(matches) > 0:
                    to_keep.extend(matches)
            if fn(idx, p):
                to_keep.append(p)
        try:
            return type(self)(to_keep)
        except TypeError:
            return self.multi_pen_class(to_keep)
    
    def index(self, idx, fn):
        fn(self[idx])
        return self
    
    def glyphs_named(self, glyph_name):
        """Pluck glyphs named `glyph_name`"""
        #return self.pfilter(lambda i, p: p.glyphName == glyph_name).pmap(lambda idx, p: mod_fn(p))
        for p in self:
            if callable(glyph_name) and glyph_name(p.glyphName):
                yield p
            elif p.glyphName == glyph_name:
                yield p
    
    def tagged(self, tag):
        """Yield all top-level pens tagged w/ `tag`"""
        for p in self:
            if p.tag() == tag:
                yield p
    
    def find(self, tag):
        matches = []
        def finder(p, a, b):
            if p.tag() == tag:
                matches.append(p)
        self.walk(finder)
        return matches
    
    def get(self, k):
        tagged = self.fft(k)
        if tagged:
            return tagged.copy()
    
    def fmmap(self, filter_fn:Callable[[int, DraftingPen], bool], map_fn:Callable[[int, DraftingPen], None]):
        for idx, p in enumerate(self._pens):
            if filter_fn(idx, p):
                map_fn(idx, p)
        return self

    def ffg(self, glyph_name):
        """(f)ind the (f)irst (g)lyph named this name"""
        return list(self.glyphs_named(glyph_name))[0]
    
    def fft(self, tag, fn=None):
        """(f)ind the (f)irst (t)agged with `tag`"""
        try:
            tagged = list(self.tagged(tag))[0]
            if fn:
                fn(tagged)
                return self
            else:
                return tagged
        except:
            if fn:
                return self
            return None
    
    def remove(self, *args):
        """remove a pen from these pens by identify, or by tag if a string is passed"""
        for k in args:
            if isinstance(k, str):
                tagged = self.fft(k)
                if tagged:
                    self._pens.remove(tagged)
            else:
                self._pens.remove(k)
        return self
    
    def mfilter(self, fn):
        """Same as `filter` but (m)utates this DATPens
        to now have only the filtered pens"""
        self._pens = self.filter(fn)
        return self
    
    def collapseonce(self):
        pens = []
        for idx, p in enumerate(self._pens):
            pens.extend(p)
        self._pens = pens
        return self
    
    def collapse(self, levels=100, onself=False):
        """AKA `flatten` in some programming contexts, though
        `flatten` is a totally different function here that flattens outlines; this function flattens nested collections into one-dimensional collections"""
        pens = []
        for idx, p in enumerate(self._pens):
            if hasattr(p, "_pens") and levels > 0:
                pens.extend(p.collapse(levels=levels-1)._pens)
            else:
                pens.append(p)
        dps = self.multi_pen_class(pens)
        if self.layered:
            dps.layered = True
        if onself:
            self._pens = dps._pens
            return self
        else:
            return dps
    
    flatten = collapse # deprecated but used in the wild
    
    def frameSet(self, th=False, tv=False):
        """All the frames of all the pens"""
        if self._frame:
            return super().frameSet(th=th, tv=tv)
        dps = self.multi_pen_class()
        for p in self._pens:
            if p._frame:
                dps.append(p.frameSet(th=th, tv=tv))
        return dps
    
    #def cast(self, _class, *args):
        #return _class(self._pens)
    #    res = _class(self, *args)
    #    res.attrs = deepcopy(self.attrs)
    #    return res


import math
from coldtype.interpolation import loopidx, interp_dict
from coldtype.time.timeable import Timing, Timeable, TimeableSet
from coldtype.time.timeline import Timeline


class LoopPhase():
    """
    Clip-like representation of a segment of a Loop, most likely retrieved from :meth:`~coldtype.time.loop.Loop.current_phase`

    * Standard ``Timeable`` available as ``.t``
    * ``.is_transition`` indicates if phase represents a transition
    """
    def __init__(self, is_transition:bool, i:int, timeable:Timeable):
        self.t = timeable
        self.i = i
        self.is_transition = is_transition
    
    def calc_state(self, states, e="eeio"):
        """
        Calculate progress of the loop phase, based on an easing fuction, ``e``

        * ``e=`` takes a mnemonic as enumerated in :func:`coldtype.time.easing.ease`
        """
        state = loopidx(states, self.t.index)
        if self.is_transition:
            next_state = loopidx(states, self.t.index+1)
            e = self.t.progress(self.i, easefn=e, to1=1).e
            state = interp_dict(e, state.copy(), next_state)
        return state


class Loop(Timeline):
    """
    Construct for quickly developing animations based on loop
    
    Can be used as a ``timeline=`` for an ``@animation``
    """
    def __init__(self, duration, transition_length, states,
        loop=1, fps=30, storyboard=[0]):

        self.timeables = []
        self.transitions = []

        super().__init__(duration, fps=fps, storyboard=storyboard)

        try:
            segment_count = len(states)
            self.states = states
        except:
            segment_count = states
            self.states = None

        idx = 0
        segment_count = int(math.floor(segment_count))
        segment_length = duration / segment_count
        segment_frames = int(math.floor(segment_length))
        leftover = duration % segment_frames

        for x in range(0, self.duration, segment_frames):
            if idx >= segment_count:
                continue

            start = x
            end = x + segment_frames
            if idx == segment_count-1:
                end += leftover
            
            self.timeables.append(Timeable(start, end, index=idx))

            if idx < segment_count - 1:
                tstart = end-transition_length
                tend = end+transition_length
                self.transitions.append(Timeable(tstart, tend, index=idx))
            
            idx += 1
        
        if loop:
            self.transitions.insert(0, Timeable(-transition_length, transition_length, index=-1))
            self.transitions.append(Timeable(self.duration-transition_length, self.duration+transition_length, index=idx-1))
    
    def jumps(self):
        js = self._jumps
        for t in self.timeables:
            js.insert(-1, t.start+int(t.duration/2))
        return js
    
    def current_clip(self, i):
        for t in self.timeables:
            if t.now(i):
                return t

    def current_transition(self, i):
        for t in self.transitions:
            if t.now(i):
                return t

    def current(self, i):
        clip = self.current_clip(i)
        transition = self.current_transition(i)
        if transition:
            return True, transition
        else:
            return False, clip

    def current_on_loop(self, i):
        if i < 0:
            i = self.duration + i
        is_transition, timeable = self.current(i)
        return is_transition, i, timeable
    
    def current_phase(self, i) -> LoopPhase:
        """
        Get current phase of the Loop as a ``LoopPhase``
        """
        return LoopPhase(*self.current_on_loop(i))
    
    def current_state(self, i, e="eeio") -> LoopPhase:
        """
        Get current state of the Loop as a dict (provided states were passed to init)
        """
        return LoopPhase(*self.current_on_loop(i)).calc_state(self.states, e)


from fontTools.pens.transformPen import TransformPen
from fontTools.pens.basePen import BasePen

from coldtype.color import Gradient, Color
from coldtype.pens.drawablepen import DrawablePenMixin

from coldtype.color import Color
from lxml import etree
import base64

from random import randint


def pointToString(pt):
    return " ".join(["{:0.1f}".format(i) for i in pt])


class SVGPathPen(BasePen):

    def __init__(self, glyphSet):
        BasePen.__init__(self, glyphSet)
        self._commands = []
        self._lastCommand = None
        self._lastX = None
        self._lastY = None

    def _handleAnchor(self):
        """
        >>> pen = SVGPathPen(None)
        >>> pen.moveTo((0, 0))
        >>> pen.moveTo((10, 10))
        >>> pen._commands
        ['M10 10']
        """
        if self._lastCommand == "M":
            self._commands.pop(-1)

    def _moveTo(self, pt):
        """
        >>> pen = SVGPathPen(None)
        >>> pen.moveTo((0, 0))
        >>> pen._commands
        ['M0 0']
        >>> pen = SVGPathPen(None)
        >>> pen.moveTo((10, 0))
        >>> pen._commands
        ['M10 0']
        >>> pen = SVGPathPen(None)
        >>> pen.moveTo((0, 10))
        >>> pen._commands
        ['M0 10']
        """
        self._handleAnchor()
        t = "M%s" % (pointToString(pt))
        self._commands.append(t)
        self._lastCommand = "M"
        self._lastX, self._lastY = pt

    def _lineTo(self, pt):
        """
        # duplicate point
        >>> pen = SVGPathPen(None)
        >>> pen.moveTo((10, 10))
        >>> pen.lineTo((10, 10))
        >>> pen._commands
        ['M10 10']
        # vertical line
        >>> pen = SVGPathPen(None)
        >>> pen.moveTo((10, 10))
        >>> pen.lineTo((10, 0))
        >>> pen._commands
        ['M10 10', 'V0']
        # horizontal line
        >>> pen = SVGPathPen(None)
        >>> pen.moveTo((10, 10))
        >>> pen.lineTo((0, 10))
        >>> pen._commands
        ['M10 10', 'H0']
        # basic
        >>> pen = SVGPathPen(None)
        >>> pen.lineTo((70, 80))
        >>> pen._commands
        ['L70 80']
        # basic following a moveto
        >>> pen = SVGPathPen(None)
        >>> pen.moveTo((0, 0))
        >>> pen.lineTo((10, 10))
        >>> pen._commands
        ['M0 0', ' 10 10']
        """
        x, y = pt
        # duplicate point
        if x == self._lastX and y == self._lastY:
            return
        # vertical line
        elif x == self._lastX:
            cmd = "V"
            pts = str(y)
        # horizontal line
        elif y == self._lastY:
            cmd = "H"
            pts = str(x)
        # previous was a moveto
        elif self._lastCommand == "M":
            cmd = None
            pts = " " + pointToString(pt)
        # basic
        else:
            cmd = "L"
            pts = pointToString(pt)
        # write the string
        t = ""
        if cmd:
            t += cmd
            self._lastCommand = cmd
        t += pts
        self._commands.append(t)
        # store for future reference
        self._lastX, self._lastY = pt

    def _curveToOne(self, pt1, pt2, pt3):
        """
        >>> pen = SVGPathPen(None)
        >>> pen.curveTo((10, 20), (30, 40), (50, 60))
        >>> pen._commands
        ['C10 20 30 40 50 60']
        """
        t = "C"
        t += pointToString(pt1) + " "
        t += pointToString(pt2) + " "
        t += pointToString(pt3)
        self._commands.append(t)
        self._lastCommand = "C"
        self._lastX, self._lastY = pt3

    def _qCurveToOne(self, pt1, pt2):
        """
        >>> pen = SVGPathPen(None)
        >>> pen.qCurveTo((10, 20), (30, 40))
        >>> pen._commands
        ['Q10 20 30 40']
        """
        assert pt2 is not None
        t = "Q"
        t += pointToString(pt1) + " "
        t += pointToString(pt2)
        self._commands.append(t)
        self._lastCommand = "Q"
        self._lastX, self._lastY = pt2

    def _closePath(self):
        """
        >>> pen = SVGPathPen(None)
        >>> pen.closePath()
        >>> pen._commands
        ['Z']
        """
        self._commands.append("Z")
        self._lastCommand = "Z"
        self._lastX = self._lastY = None

    def _endPath(self):
        """
        >>> pen = SVGPathPen(None)
        >>> pen.endPath()
        >>> pen._commands
        ['Z']
        """
        self._closePath()
        self._lastCommand = None
        self._lastX = self._lastY = None

    def getCommands(self):
        return "".join(self._commands)


class SVGPen(DrawablePenMixin, SVGPathPen):
    def __init__(self, dat, h):
        super().__init__(None)
        self.defs = []
        self.uses = []
        self.dat = dat
        self.h = h
        tp = TransformPen(self, (1, 0, 0, -1, 0, h))
        dat.round_to(0.1).replay(tp)
    
    def _endPath(self):
        """
        >>> pen = SVGPathPen(None)
        >>> pen.endPath()
        >>> pen._commands
        ['Z']
        """
        #self._closePath()
        self._lastCommand = None
        self._lastX = self._lastY = None
    
    def fill(self, color):
        if color:
            if isinstance(color, Gradient):
                self.path.set("fill", f"url('#{self.gradient(color)}')")
            elif isinstance(color, Color):
                self.path.set("fill", self.rgba(color))
        else:
            self.path.set("fill", "transparent")
    
    def stroke(self, weight=1, color=None, dash=None):
        self.path.set("stroke-width", str(weight))
        if dash:
            self.path.set("stroke-dasharray", " ".join([str(x) for x in dash]))
        if color:
            if isinstance(color, Gradient):
                self.path.set("stroke", f"url('#{self.gradient(color)}')")
            elif isinstance(color, Color):
                self.path.set("stroke", self.rgba(color))
        else:
            self.path.set("stroke-width", 0)
            self.path.set("stroke", "transparent")
    
    def rgba(self, color):
        r, g, b, a = color.ints()
        return f"rgba({r}, {g}, {b}, {a})"
    
    def rect(self, rect):
        r = etree.Element("rect")
        fr = rect.flip(self.h)
        r.set("x", str(fr.x))
        r.set("y", str(fr.y))
        r.set("width", str(fr.w))
        r.set("height", str(fr.h))
        return r
    
    def shadow(self, clip=None, radius=10, alpha=0.3, color=Color.from_rgb(0,0,0,1)):
        hsh = str(hash(self.getCommands())) + str(randint(0, 1000000))
        f = etree.Element("filter")
        f.set("x", "0")
        f.set("y", "0")
        f.set("width", "1000%")
        f.set("height", "1000%")
        f.set("x", "-500%")
        f.set("y", "-500%")
        f.set("id", f"shadow-{hsh}")
        fe = etree.Element("feDropShadow")
        fe.set("dx", "0")
        fe.set("dy", "0")
        fe.set("stdDeviation", str(radius))
        #fe.set("slope", str(alpha))
        fe.set("flood-color", self.rgba(color))
        fe.set("flood-opacity", str(alpha))
        f.append(fe)
        self.defs.append(f)
        if clip:
            cp = etree.Element("clipPath")
            cp.set("id", f"clip-path_{hsh}")
            cp.append(self.rect(clip))
            self.defs.append(cp)
        #self.path.set("fill", "rgba(0, 0, 0, 0.3)")
        self.path.set("filter", f"url(#{f.get('id')})")
        if clip:
            self.path.set("clip-path", f"url(#clip-path_{hsh})")

    def gradient(self, gradient):
        lg = etree.Element("linearGradient")
        lg.set("id", f"gradient-{hash(self.getCommands())}-{randint(0, 100000)}")
        if gradient.stops[1][1].x == gradient.stops[0][1].x:
            lg.set("gradientTransform", "rotate(90)")
        s1 = etree.Element("stop", offset="0%")
        s1.set("stop-color", self.rgba(gradient.stops[0][0]))
        s2 = etree.Element("stop", offset="100%")
        s2.set("stop-color", self.rgba(gradient.stops[1][0]))
        lg.append(s1)
        lg.append(s2)
        self.defs.append(lg)
        return lg.get("id")
    
    def image(self, src=None, opacity=None, rect=None):
        if True:
            src = base64.b64encode(open(src, "rb").read()).decode('utf-8')
        hsh = str(hash(self.getCommands())) + str(randint(0, 100000))
        img = etree.Element("image")
        img.set("x", str(rect.x or 0))
        img.set("y", str(rect.y or 0))
        img.set("width", str(rect.w or 100))
        img.set("height", str(rect.h or 100))
        img.set("opacity", str(opacity))
        img.set("image-href", f"data:image/png;base64,{src}")
        pattern = etree.Element("pattern")
        pattern.set("x", img.get("x"))
        pattern.set("y", img.get("y"))
        pattern.set("width", img.get("width"))
        pattern.set("height", img.get("height"))
        pattern.set("patternUnits", "userSpaceOnUse")
        pattern.set("id", f"pattern-{hsh}")
        pattern.append(img)
        self.defs.append(pattern)
        self.path.set("fill", f"url(#pattern-{hsh})")
    
    def asSVG(self, style=None):
        self.path = etree.Element("path")
        for attrs, attr in self.findStyledAttrs(style):
            self.applyDATAttribute(attrs, attr)
        self.path.set("d", self.getCommands())
        tag = self.dat.tag()
        if tag:
            self.path.set("data-tag", tag)
        g = etree.Element("g")
        defs = etree.Element("defs")
        for d in self.defs:
            defs.append(d)
        g.append(defs)
        g.append(self.path)
        for u in self.uses:
            g.append(u)
        return g
    
    def Composite(pens, rect, offset=False, style=None, viewBox=False):
        docroot = etree.Element("svg")
        docroot.set("xmlns", "http://www.w3.org/2000/svg")
        if not viewBox:
            docroot.set("width", str(rect.w))
            docroot.set("height", str(rect.h))
        else:
            docroot.set("viewBox", f"0 0 {rect.w} {rect.h}")
            docroot.set("width", "100%")
        if offset:
            docroot.set("style", f"left:{rect.x}px;bottom:{rect.y}px;")
        
        for pen in SVGPen.FindPens(pens):
            sp = SVGPen(pen, rect.h)
            docroot.append(sp.asSVG(style=style))
        
        return etree.tostring(docroot, pretty_print=True).decode("utf-8").replace("image-href", "xlink:href")


import math
from coldtype.geometry.geometrical import Geometrical
from coldtype.interpolation import norm
from coldtype.geometry.primitives import polar_coord, line_intersection, calc_angle, calc_vector


class Point(Geometrical):
    """Representation of a point (x,y), indexable"""
    def __init__(self, *points):
        if len(points) == 2:
            self.x = points[0]
            self.y = points[1]
        else:
            point = points[0]
            try:
                x, y = point
                self.x = x
                self.y = y
            except TypeError:
                try:
                    self.x = point.x
                    self.y = point.y
                except AttributeError:
                    self.x = 0
                    self.y = 0
    
    __hash__ = object.__hash__

    def Z():
        return Point([0, 0])

    def from_obj(obj):
        p = Point((0, 0))
        try:
            p.x = obj.x
            p.y = obj.y
        except:
            pass
        return p

    def offset(self, dx, dy):
        "Offset by dx, dy"
        return Point((self.x + dx, self.y + dy))
    
    def offset_x(self, dx):
        return self.offset(dx, 0)
    
    def offset_y(self, dy):
        return self.offset(0, dy)
    
    o = offset

    def rect(self, w, h):
        "Create a rect from this point as center, with w and h dimensions provided"
        from coldtype.geometry.rect import Rect
        return Rect((self.x-w/2, self.y-h/2, w, h))

    def xy(self):
        "As a tuple"
        return self.x, self.y
    
    def round(self):
        """round the values in the point to the nearest integer"""
        return Point([int(round(n)) for n in self])
    
    def round_to(self, to=10):
        """round the values in the point to the nearest integer multiple"""
        return Point([int(round(n/to)*to) for n in self.xy()])
    
    def inside(self, rect):
        mnx, mny, mxx, mxy = rect.mnmnmxmx()
        if mnx <= self.x <= mxx and mny <= self.y <= mxy:
            return True
        else:
            return False

    def flip(self, frame):
        return Point((self.x, frame.h - self.y))

    def flipSelf(self, frame):
        x, y = self.flip(frame)
        self.x = x
        self.y = y
    
    def scale(self, x, y=None):
        if not y:
            y = x
        return Point((self.x * x, self.y * y))
    
    def join(self, other):
        from coldtype.geometry.line import Line
        return Line(self, other)
    
    def interp(self, v, other):
        """Interpolate with another point"""
        sx, sy = self
        ox, oy = other
        if not isinstance(v, float) and v != 1:
            v = v/100
        return Point((norm(v, sx, ox), norm(v, sy, oy)))
    
    def i(self, *args):
        other = args
        if isinstance(other[0], Point):
            return self.interp(0.5, other[0])
        else:
            x, pt = other
            return self.interp(x, pt)
    
    def project(self, angle, dist):
        dx, dy = polar_coord((0, 0), math.radians(angle), dist)
        return self.offset(dx, dy)
    
    def project_to(self, angle, line):
        ray = [self, self.project(angle, 1000)]
        return Point(line_intersection(ray, line))
    
    def cdist(self, other):
        vec = calc_vector(self, other)
        ang = calc_angle(self, other)
        dist = math.sqrt(math.pow(vec[0], 2) + math.pow(vec[1], 2))
        return dist, math.degrees(ang)
    
    def noop(self, *args, **kwargs):
        return self
    
    def __eq__(self, o):
        try:
            return all([self.x == o.x, self.y == o.y])
        except:
            return False

    def __repr__(self):
        return "Point" + str(self.xy())

    def __getitem__(self, key):
        return self.xy()[key]
    
    def __len__(self):
        return 2

    def __setitem__(self, key, value):
        if key == 0:
            self.x = value
        elif key == 1:
            self.y = value
        else:
            raise IndexError(
                "Invalid index for point assignment, must be 0 or 1")
    
    def __floordiv__(self, other):
        return self.offset(0, other)
    
    def __truediv__(self, other):
        return self.offset(other, 0)
    
    def setx(self, x):
        return Point([x, self.y])
    
    def __mul__(self, other):
        return self.setx(other)
    
    def sety(self, y):
        return Point([self.x, y])
    
    def __matmul__(self, other):
        return self.sety(other)
    
    def __or__(self, other):
        from coldtype.geometry.line import Line
        return Line(self, other)
    
    #def as3d(self):
    #    return Point3D(self)
    
    def reverse(self):
        return Point(self.y, self.x)


# class Point3D(Point):
#     def __init__(self, p):
#         super().__init__(p)
#         self.z = 0
    
#     def __len__(self):
#         return 3
    
#     def __getitem__(self, idx):
#         if idx == 2:
#             return self.z
#         return super().__getitem__(idx)


import os
import tempfile

from .dsCompiler import compileDSToPath
from .ufoCompiler import compileUFOToPath as ufo_compileUFOToPath
from .ttxCompiler import compileTTXToPath

def compileUFOToPath(ufoPath, ttPath, outputWriter):
    return ufo_compileUFOToPath(os.fspath(ufoPath), os.fspath(ttPath))

def compileUFOToBytes(ufoPath, outputWriter):
    with tempfile.NamedTemporaryFile(prefix="fontgoggles_temp", suffix=".ttf") as tmp:
        compileUFOToPath(ufoPath, tmp.name, outputWriter)
        with open(tmp.name, "rb") as f:
            fontData = f.read()
            if not fontData:
                fontData = None
    return fontData


def compileDSToBytes(dsPath, ttFolder, outputWriter):
    with tempfile.NamedTemporaryFile(prefix="fontgoggles_temp", suffix=".ttf") as tmp:
        compileDSToPath(dsPath, ttFolder, tmp.name)
        with open(tmp.name, "rb") as f:
            fontData = f.read()
            if not fontData:
                fontData = None
    return fontData


def compileTTXToBytes(ttxPath, outputWriter):
    with tempfile.NamedTemporaryFile(prefix="fontgoggles_temp", suffix=".ttf") as tmp:
        compileTTXToPath(ttxPath, tmp.name)
        with open(tmp.name, "rb") as f:
            fontData = f.read()
            if not fontData:
                fontData = None
    return fontData


from fontTools.pens.transformPen import TransformPen
from fontTools.pens.basePen import BasePen
from pathlib import Path

try:
    import cairo
except:
    print(">>> No cairo installation found!")


from coldtype.pens.drawablepen import DrawablePenMixin
from coldtype.color import Color, Gradient
from coldtype.beziers import raise_quadratic


class CairoPen(DrawablePenMixin, BasePen):
    def __init__(self, dat, h, ctx, style=None):
        super().__init__(None)
        self.dat = dat
        self.h = h
        self.ctx = ctx
        self._value = []
        tp = TransformPen(self, (1, 0, 0, -1, 0, h))
        
        attrs = list(self.findStyledAttrs(style))
        methods = [a[0] for a in attrs]

        if True or "shadow" not in methods:
            for attrs, attr in self.findStyledAttrs(style):
                method, *args = attr
                self.ctx.save()
                if method in ["fill", "stroke"]:
                    dat.replay(tp)
                self.applyDATAttribute(attrs, attr)
                self.ctx.restore()

    def _moveTo(self, p):
        self.ctx.move_to(p[0], p[1])
        self._value.append(p)

    def _lineTo(self, p):
        self.ctx.line_to(p[0], p[1])
        self._value.append(p)

    def _curveToOne(self, p1, p2, p3):
        self.ctx.curve_to(p1[0], p1[1], p2[0], p2[1], p3[0], p3[1])
        self._value.extend([p1, p2, p3])

    def _qCurveToOne(self, p1, p2):
        start = self._value[-1]
        q1, q2, q3 = raise_quadratic(start, (p1[0], p1[1]), (p2[0], p2[1]))
        self.ctx.curve_to(q1[0], q1[1], q2[0], q2[1], q3[0], q3[1])
        self._value.extend([q1, q2, q3])

    def _closePath(self):
        self.ctx.close_path()
    
    def fill(self, color=None):
        if color:
            if isinstance(color, Gradient):
                self.gradient(color)
            else:
                self.ctx.set_source_rgba(color.r, color.g, color.b, color.a)
            self.ctx.fill()
    
    def stroke(self, weight=1, color=None, dash=None):
        self.ctx.set_source_rgba(color.r, color.g, color.b, color.a)
        self.ctx.set_line_width(weight)
        self.ctx.stroke()
    
    def gradient(self, gradient):
        pat = cairo.LinearGradient(*[p for s in reversed(gradient.stops) for p in s[1]])
        for idx, stop in enumerate(gradient.stops):
            c = stop[0]
            pat.add_color_stop_rgba(idx, c.r, c.g, c.b, c.a)
        self.ctx.set_source(pat)
    
    def image(self, src=None, opacity=None, rect=None):
        image_surface = cairo.ImageSurface.create_from_png(src)
        pattern = cairo.SurfacePattern(image_surface)
        pattern.set_extend(cairo.Extend.REPEAT)
        if rect:
            self.ctx.scale(rect.h/self.h/2, rect.h/self.h/2)
        else:
            self.ctx.scale(0.5, 0.5)
        #self.ctx.translate(left, top)
        self.ctx.set_source(pattern)
        #self.ctx.set_source_surface(pattern)
        self.ctx.paint_with_alpha(opacity)
    
    def shadow(self, clip=None, radius=10, alpha=0.3, color=Color.from_rgb(1,0,0,1)):
        pass
    
    def Composite(pens, rect, image_path, save=True, style=None):
        ip = Path(image_path)
        if ip.suffix == ".png":
            surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, int(rect.w*2), int(rect.h*2))
            ctx = cairo.Context(surface)
            ctx.scale(2, 2)
            for pen in CairoPen.FindPens(pens):
                CairoPen(pen, rect.h, ctx, style=style)
            if save:
                surface.write_to_png(image_path)
            else:
                print("Should write to base64 and return — not yet supported")
        elif ip.suffix == ".pdf":
            surface = cairo.PDFSurface(str(ip), int(rect.w), int(rect.h))
            #print(surface)
            ctx = cairo.Context(surface)
            #ctx.scale(2, 2)
            for pen in CairoPen.FindPens(pens):
                CairoPen(pen, rect.h, ctx, style=style)
        else:
            raise Exception(f"CairoPen cannot print to format {ip.suffix}")


import itertools
from .segmenting import textSegments


alignments = dict(LTR="left", RTL="right", TTB="top", BTT="bottom")


class TextInfo:

    def __init__(self, text):
        self.text = text
        self.shouldApplyBiDi = True  # More like .shouldApplyBiDiAndSegmentation but that's looong
        self.directionOverride = None
        self.scriptOverride = None
        self.languageOverride = None

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, text):
        self._text = text
        self._segments, self.baseLevel = textSegments(text)
        self.reorderedSegments = self._getReorderedSegments()

        toBiDi = {}
        fromBiDi = {}
        afterIndex = 0
        for segmentText, segmentScript, segmentBiDiLevel, firstCluster in self.reorderedSegments:
            charIndices = [beforeIndex
                           for beforeIndex in range(firstCluster, firstCluster + len(segmentText))]
            if segmentBiDiLevel % 2:
                charIndices = reversed(charIndices)

            for beforeIndex in charIndices:
                toBiDi[beforeIndex] = afterIndex
                fromBiDi[afterIndex] = beforeIndex
                afterIndex += 1

        assert len(toBiDi) == len(text)
        assert len(fromBiDi) == len(text)
        self._toBiDi = toBiDi
        self._fromBiDi = fromBiDi

    @property
    def segments(self):
        if self.shouldApplyBiDi:
            return self.reorderedSegments
        else:
            return [(self._text, None, None, 0)]

    def _getReorderedSegments(self):
        segments = []
        isRTL = self.baseLevel % 2
        for value, sub in itertools.groupby(self._segments, key=lambda item: item[2] % 2):
            if isRTL == value:
                segments.extend(sub)
            else:
                segments.extend(reversed(list(sub)))
        if isRTL:
            segments = list(reversed(segments))
        assert len(segments) == len(self._segments)
        return segments

    def mapToBiDi(self, charIndices):
        toBiDi = self._toBiDi
        return [toBiDi[charIndex] for charIndex in charIndices]

    def mapFromBiDi(self, charIndices):
        fromBiDi = self._fromBiDi
        return [fromBiDi[charIndex] for charIndex in charIndices]

    @property
    def baseDirection(self):
        return ("L", "R")[self.baseLevel % 2]

    @property
    def direction(self):
        if self.directionOverride is not None:
            return self.directionOverride
        else:
            return ("LTR", "RTL")[self.baseLevel % 2]

    @property
    def suggestedAlignment(self):
        alignments = dict(LTR="left", RTL="right", TTB="top", BTT="bottom")
        return alignments[self.direction]
